# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import json
import logging
import os
import sys
from datetime import datetime
from heapq import heappop, heappush
from typing import Any, Optional

from rich.console import Console
from rich.panel import Panel
from rich.text import Text

from .alerts import AlertLevel, AlertStore
from .config import settings
from .indicators import Indicators
from .module import EncryptedBackupError, MVTModule, run_module, save_timeline
from .module_loader import (
    ModuleOrigin,
    get_module_logger,
    get_module_origin,
    module_supports_command,
)
from .module_types import ModuleTimeline, URLResult
from .utils import (
    CustomJSONEncoder,
    convert_datetime_to_iso,
    generate_hashes_from_path,
    get_sha256_from_file_path,
)
from .version import MVT_VERSION


class Command:
    def __init__(
        self,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        ioc_files: Optional[list] = None,
        iocs: Optional[Indicators] = None,
        module_name: Optional[str] = None,
        serial: Optional[str] = None,
        module_options: Optional[dict] = None,
        hashes: Optional[bool] = False,
        sub_command: Optional[bool] = False,
        log: logging.Logger = logging.getLogger(__name__),
        disable_version_check: bool = False,
        disable_indicator_check: bool = False,
        custom_modules: Optional[list[type[MVTModule]]] = None,
    ) -> None:
        self.name = ""
        self.platform = ""
        self.modules: list[type[MVTModule]] = []
        self.custom_modules = custom_modules if custom_modules else []

        self.target_path = target_path
        self.results_path = results_path
        self.ioc_files = ioc_files if ioc_files else []
        self.module_name = module_name
        self.serial = serial
        self.log = log
        self.sub_command = sub_command
        self.disable_version_check = disable_version_check
        self.disable_indicator_check = disable_indicator_check

        # This dictionary can contain options that will be passed down from
        # the Command to all modules. This can for example be used to pass
        # down a password to decrypt a backup or flags which are need by some modules.
        self.module_options = module_options if module_options else {}

        # This dictionary maps the modules which were replaced by a module
        # declaring `replaces` to the module which took their place.
        self.module_replacements: dict[type[MVTModule], type[MVTModule]] = {}

        # This list will contain all executed modules.
        # We can use this to reference e.g. self.executed[0].results.
        self.executed: list[MVTModule] = []
        self.hashes = hashes
        self.hash_values: list[dict[str, Any]] = []
        self.timeline: ModuleTimeline = []
        self.url_results: list[URLResult] = []

        # Load IOCs
        self._create_storage()
        self._setup_logging()

        if iocs is not None:
            self.iocs = iocs
        else:
            self.iocs = Indicators(self.log)
            self.iocs.load_indicators_files(self.ioc_files)

        self.alertstore = AlertStore()

    def _create_storage(self) -> None:
        if self.results_path and not os.path.exists(self.results_path):
            try:
                os.makedirs(self.results_path)
            except Exception as exc:
                self.log.fatal(
                    "Unable to create output folder %s: %s", self.results_path, exc
                )
                sys.exit(1)

    def _setup_logging(self):
        if not self.results_path:
            return

        logger = logging.getLogger("mvt")
        file_handler = logging.FileHandler(
            os.path.join(self.results_path, "command.log")
        )
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        # MVT can be run in a loop.
        # Old file handlers stick around in subsequent loops.
        # Remove any existing logging.FileHandler instances.
        for handler in logger.handlers:
            if isinstance(handler, logging.FileHandler):
                logger.removeHandler(handler)

        # And finally add the new one.
        logger.addHandler(file_handler)

    def _store_timeline(self) -> None:
        if not self.results_path:
            return

        # We use local timestamps in the timeline on Android as many
        # logs do not contain timezone information.
        if type(self).__name__.startswith("CmdAndroid"):
            is_utc = False
        else:
            is_utc = True

        if len(self.timeline) > 0:
            save_timeline(
                self.timeline,
                os.path.join(self.results_path, "timeline.csv"),
                is_utc=is_utc,
            )

    def _store_alerts(self) -> None:
        if not self.results_path:
            return

        alerts = self.alertstore.as_json()
        if not alerts:
            return

        alerts_path = os.path.join(self.results_path, "alerts.json")
        with open(alerts_path, "w+", encoding="utf-8") as handle:
            json.dump(alerts, handle, indent=4, cls=CustomJSONEncoder)

    def _store_urls(self) -> None:
        if not self.results_path or not self.url_results:
            return

        urls_path = os.path.join(self.results_path, "urls.json")
        with open(urls_path, "w", encoding="utf-8") as handle:
            json.dump(self.url_results, handle, indent=4, cls=CustomJSONEncoder)

    def _store_alerts_timeline(self) -> None:
        if not self.results_path:
            return

        alerts_timeline_path = os.path.join(self.results_path, "alerts_timeline.csv")
        self.alertstore.save_timeline(alerts_timeline_path)

    def _store_info(self) -> None:
        if not self.results_path:
            return

        target_path: Optional[str] = None
        if self.target_path:
            target_path = os.path.abspath(self.target_path)

        info: dict[str, Any] = {
            "target_path": target_path,
            "mvt_version": MVT_VERSION,
            "date": convert_datetime_to_iso(datetime.now()),
            "ioc_files": [],
            "hashes": [],
        }

        for coll in self.iocs.ioc_collections:
            ioc_file_path = coll.get("stix2_file_path", "")
            if ioc_file_path and ioc_file_path not in info["ioc_files"]:
                info["ioc_files"].append(ioc_file_path)

        if self.target_path and (settings.HASH_FILES or self.hashes):
            self.generate_hashes()

        info["hashes"] = self.hash_values

        info_path = os.path.join(self.results_path, "info.json")
        with open(info_path, "w+", encoding="utf-8") as handle:
            json.dump(info, handle, indent=4)

        if self.target_path and (settings.HASH_FILES or self.hashes):
            info_hash = get_sha256_from_file_path(info_path)
            self.log.info('Reference hash of the info.json file: "%s"', info_hash)

    def generate_hashes(self) -> None:
        """
        Compute hashes for files in the target_path
        """
        if not self.target_path:
            return

        for file in generate_hashes_from_path(self.target_path, self.log):
            self.hash_values.append(file)

    @staticmethod
    def _modules_by_origin(
        modules: list[type[MVTModule]],
    ) -> dict[ModuleOrigin, list[str]]:
        origins: dict[ModuleOrigin, list[str]] = {}
        for module in modules:
            origins.setdefault(get_module_origin(module), []).append(module.__name__)
        return origins

    def list_modules(self) -> None:
        self.log.info("Following is the list of available %s modules:", self.name)
        for origin, module_names in self._modules_by_origin(
            self._available_modules()
        ).items():
            self.log.info(
                " - Modules from %s: %s", origin.label, ", ".join(module_names)
            )

    def _log_loaded_modules(self, modules: list[type[MVTModule]]) -> None:
        """Record the loaded modules and their origins for auditability."""
        for origin, module_names in self._modules_by_origin(modules).items():
            self.log.info(
                "Loaded %d %s modules from %s: %s",
                len(module_names),
                self.name,
                origin.label,
                ", ".join(module_names),
            )

    def _available_modules(self) -> list[type[MVTModule]]:
        modules = list(self.modules)
        modules.extend(
            module
            for module in self.custom_modules
            if module_supports_command(module, self.platform, self.name)
        )

        deduplicated = []
        for module in modules:
            if module not in deduplicated:
                deduplicated.append(module)

        available = self._apply_replacements(deduplicated)
        self._warn_about_slug_collisions(available)

        return available

    def _warn_about_slug_collisions(self, modules: list[type[MVTModule]]) -> None:
        """Report modules writing their results to the same file.

        Results are stored in a file named after the module slug, so two
        modules sharing one silently overwrite each other. Replacements are
        already resolved here, so a module deliberately taking over the slug
        of the module it replaces is not reported: the module it replaced is
        no longer part of the run.
        """
        modules_by_slug: dict[str, type[MVTModule]] = {}
        for module in modules:
            slug = module.get_slug()
            first = modules_by_slug.setdefault(slug, module)
            if first is module:
                continue

            self.log.warning(
                "Modules %s from %s and %s from %s both use the slug %s. If "
                "both run, whichever runs last overwrites the results of the "
                "other in %s.json.",
                first.__name__,
                get_module_origin(first).label,
                module.__name__,
                get_module_origin(module).label,
                slug,
                slug,
            )

    def _declared_replacements(
        self, modules: list[type[MVTModule]]
    ) -> dict[type[MVTModule], type[MVTModule]]:
        """Return the replacements declared by the given modules."""
        replacements: dict[type[MVTModule], type[MVTModule]] = {}
        for module in modules:
            replaced = module.replaces
            if replaced is None or replaced is module:
                continue

            if not module.enabled:
                # Replacing a module must not disable it, as a disabled
                # replacement never runs in its place.
                self.log.debug(
                    "Module %s is disabled and does not replace module %s.",
                    module.__name__,
                    replaced.__name__,
                )
                continue

            if replaced not in modules:
                # A module can support several commands while the module it
                # replaces is only available in some of them.
                self.log.debug(
                    "Module %s replaces module %s, which is not available "
                    "for the %s command.",
                    module.__name__,
                    replaced.__name__,
                    self.name,
                )
                continue

            if replaced in replacements:
                self.log.warning(
                    "Modules %s and %s both replace module %s. Both of them "
                    "will run, %s will not, and modules depending on %s will "
                    "use the results of %s. Replacements which share the slug "
                    "of %s overwrite each other's results file.",
                    replacements[replaced].__name__,
                    module.__name__,
                    replaced.__name__,
                    replaced.__name__,
                    replaced.__name__,
                    replacements[replaced].__name__,
                    replaced.__name__,
                )
                continue

            replacements[replaced] = module

        return replacements

    def _drop_replacement_cycles(
        self, replacements: dict[type[MVTModule], type[MVTModule]]
    ) -> None:
        """Undo the replacements between modules which replace each other."""
        cyclic: set[type[MVTModule]] = set()
        for replaced in replacements:
            walked: list[type[MVTModule]] = []
            module = replaced
            while module in replacements and module not in cyclic:
                if module in walked:
                    cyclic.update(walked[walked.index(module) :])
                    break
                walked.append(module)
                module = replacements[module]

        if not cyclic:
            return

        self.log.warning(
            "Modules %s replace each other in a cycle. None of them replaces "
            "anything and all of them will run.",
            ", ".join(sorted(module.__name__ for module in cyclic)),
        )
        for module in cyclic:
            replacements.pop(module, None)

    def _log_replacement(
        self,
        replaced: type[MVTModule],
        module: type[MVTModule],
        replacement: type[MVTModule],
    ) -> None:
        """Report an applied replacement, and any problem with it."""
        if not issubclass(module, replaced):
            self.log.warning(
                "Module %s replaces module %s but is not a subclass of it. "
                "Its results might not be compatible with what modules "
                "depending on %s expect.",
                module.__name__,
                replaced.__name__,
                replaced.__name__,
            )

        if replaced in module.dependencies:
            self.log.warning(
                "Module %s depends on module %s, which it also replaces. The "
                "dependency cannot be satisfied: a replacement has to produce "
                "that data itself.",
                module.__name__,
                replaced.__name__,
            )

        self.log.info(
            "Module %s from %s replaces module %s from %s.",
            replacement.__name__,
            get_module_origin(replacement).label,
            replaced.__name__,
            get_module_origin(replaced).label,
        )

    def _apply_replacements(
        self, modules: list[type[MVTModule]]
    ) -> list[type[MVTModule]]:
        """Drop the modules superseded by a module declaring `replaces`."""
        declared = self._declared_replacements(modules)
        self._drop_replacement_cycles(declared)

        replacements: dict[type[MVTModule], type[MVTModule]] = {}
        for replaced, module in declared.items():
            # Follow chains of replacements, so that a dependency on a
            # replaced module always resolves to a module which is part of
            # the run.
            replacement = module
            while replacement in declared:
                replacement = declared[replacement]

            # Only the replacements which are applied are reported, so that
            # the record matches the modules which actually run.
            self._log_replacement(replaced, module, replacement)
            replacements[replaced] = replacement

        self.module_replacements = replacements
        return [module for module in modules if module not in replacements]

    def init(self) -> None:
        raise NotImplementedError

    def module_init(self, module: MVTModule) -> None:
        raise NotImplementedError

    def finish(self) -> None:
        raise NotImplementedError

    def show_alerts_brief(self) -> None:
        console = Console()

        message = Text()
        for i, level in enumerate(AlertLevel):
            message.append(
                f"MVT produced {self.alertstore.count(level)} {level.name} alerts."
            )
            if i < len(AlertLevel) - 1:
                message.append("\n")

        panel = Panel(
            message, title="ALERTS", style="sandy_brown", border_style="sandy_brown"
        )
        console.print("")
        console.print(panel)

    def show_disable_adb_warning(self) -> None:
        console = Console()
        message = Text(
            "Please disable Developer Options and ADB (Android Debug Bridge) on the device once finished with the acquisition. "
            "ADB is a powerful tool which can allow unauthorized access to the device."
        )
        panel = Panel(message, title="NOTE", style="yellow", border_style="yellow")
        console.print("")
        console.print(panel)

    def show_support_message(self) -> None:
        console = Console()
        message = Text()

        support_message = "Please seek reputable expert help if you have serious concerns about a possible spyware attack. Such support is available to human rights defenders and civil society through Amnesty International's Security Lab at https://securitylab.amnesty.org/get-help/?c=mvt"
        if (
            self.alertstore.count(AlertLevel.HIGH) > 0
            or self.alertstore.count(AlertLevel.CRITICAL) > 0
        ):
            message.append(
                f"MVT produced HIGH or CRITICAL alerts. Only expert review can confirm if the detected indicators are signs of an attack.\n\n{support_message}",
            )
            panel = Panel(message, title="WARNING", style="red", border_style="red")
        else:
            message.append(
                f"The lack of severe alerts does not equate to a clean bill of health.\n\n{support_message}",
            )
            panel = Panel(message, title="NOTE", style="yellow", border_style="yellow")

        console.print("")
        console.print(panel)

    def _module_dependencies(
        self, module: type[MVTModule]
    ) -> list[tuple[type[MVTModule], type[MVTModule]]]:
        """Return the (declared, resolved) dependencies of a module.

        A dependency on a module which was replaced is resolved to the module
        which took its place. A module which replaces one of its own
        dependencies is not made to depend on itself, while a module which
        declares itself as a dependency is left alone and still fails the
        circular dependency check.
        """
        dependencies = []
        for dependency in module.dependencies:
            resolved = self.module_replacements.get(dependency, dependency)
            if resolved is module and dependency is not module:
                continue
            dependencies.append((dependency, resolved))

        return dependencies

    @staticmethod
    def _dependency_name(
        declared: type[MVTModule], resolved: type[MVTModule]
    ) -> str:
        """Return how a dependency is named in the messages about it.

        A dependency is named as the module which declared it wrote it, and,
        when that module was replaced, as the module which runs in its place.
        """
        if declared is resolved:
            return declared.__name__

        return f"{resolved.__name__} (replacing module {declared.__name__})"

    def _selected_modules(
        self, modules: list[type[MVTModule]]
    ) -> Optional[list[type[MVTModule]]]:
        """Return the modules explicitly requested, or all the enabled ones.

        Returns None when a module was requested by name and no module of
        that name can be run.
        """
        if not self.module_name:
            return [module for module in modules if module.enabled]

        selected = [
            module for module in modules if module.__name__ == self.module_name
        ]

        # A module replacing another one does not have to keep its name, so
        # the name of a replaced module selects its replacement.
        if not selected:
            for replaced, replacement in self.module_replacements.items():
                if replaced.__name__ != self.module_name or replacement in selected:
                    continue
                self.log.info(
                    "Module %s was replaced by module %s, which is run "
                    "in its place.",
                    replaced.__name__,
                    replacement.__name__,
                )
                selected.append(replacement)

        if not selected:
            self.log.warning(
                "No module named %s is available for the %s command. "
                "No modules will be run.",
                self.module_name,
                self.name,
            )
            return None

        return selected

    def _skipped_modules(
        self,
        required: list[type[MVTModule]],
        module_indexes: dict[type[MVTModule], int],
    ) -> dict[type[MVTModule], tuple[type[MVTModule], type[MVTModule]]]:
        """Return the modules to drop because a dependency is unavailable.

        A module declaring a dependency this command cannot provide is unable
        to run, and so is every module depending on it. Dropping only those
        keeps a single wrong declaration - in a module scoped to several
        commands, for example - from silencing an entire analysis.

        The returned mapping gives, for each skipped module, the module which
        is missing a dependency and the dependency it is missing.
        """
        skipped: dict[type[MVTModule], tuple[type[MVTModule], type[MVTModule]]] = {}
        # Skipping the one module a run was asked for leaves nothing to run,
        # which the caller reports instead.
        remainder = (
            "" if self.module_name else " The rest of the analysis will still run."
        )

        # Skipping one module can skip the modules depending on it, which the
        # pass over the module list may already have gone past, so repeat the
        # pass until nothing changes.
        changed = True
        while changed:
            changed = False
            for module in required:
                if module in skipped:
                    continue

                for declared, dependency in self._module_dependencies(module):
                    if dependency not in module_indexes:
                        # A replaced dependency always resolves to a module of
                        # this command, so an unavailable one is always the
                        # module class the author declared.
                        skipped[module] = (module, declared)
                        changed = True
                        self.log.warning(
                            "Module %s will be SKIPPED: it depends on module "
                            "%s, which is not available in this command.%s",
                            module.__name__,
                            declared.__name__,
                            remainder,
                        )
                        break

                    if dependency in skipped:
                        root, missing = skipped[dependency]
                        skipped[module] = (root, missing)
                        changed = True
                        if dependency is root:
                            self.log.warning(
                                "Module %s will be SKIPPED: it depends on "
                                "module %s, itself skipped for depending on "
                                "unavailable module %s.%s",
                                module.__name__,
                                self._dependency_name(declared, dependency),
                                missing.__name__,
                                remainder,
                            )
                        else:
                            self.log.warning(
                                "Module %s will be SKIPPED: it depends on "
                                "skipped module %s, in a chain starting at "
                                "module %s, which depends on unavailable "
                                "module %s.%s",
                                module.__name__,
                                self._dependency_name(declared, dependency),
                                root.__name__,
                                missing.__name__,
                                remainder,
                            )
                        break

        return skipped

    def _ordered_modules(self) -> Optional[list[type[MVTModule]]]:
        """Return enabled modules in stable topological order."""
        modules = self._available_modules()
        module_indexes = {module: index for index, module in enumerate(modules)}

        selected = self._selected_modules(modules)
        if selected is None:
            return None

        required: set[type[MVTModule]] = set()
        pending = list(selected)
        while pending:
            module = pending.pop()
            if module in required:
                continue
            required.add(module)
            for _, dependency in self._module_dependencies(module):
                # Unavailable dependencies are reported by _skipped_modules().
                if dependency in module_indexes:
                    pending.append(dependency)

        ordered_required = sorted(required, key=lambda module: module_indexes[module])
        skipped = self._skipped_modules(ordered_required, module_indexes)
        runnable = [module for module in ordered_required if module not in skipped]
        if skipped and not runnable:
            self.log.warning(
                "Every selected module was skipped for an unavailable "
                "dependency. No modules will be run."
            )
            return None

        dependents: dict[type[MVTModule], list[type[MVTModule]]] = {
            module: [] for module in runnable
        }
        indegree = {module: 0 for module in runnable}
        for module in runnable:
            for _, dependency in self._module_dependencies(module):
                if dependency not in indegree:
                    continue
                dependents[dependency].append(module)
                indegree[module] += 1

        ready: list[tuple[int, type[MVTModule]]] = []
        for module, count in indegree.items():
            if count == 0:
                heappush(ready, (module_indexes[module], module))

        ordered = []
        while ready:
            _, module = heappop(ready)
            ordered.append(module)
            for dependent in dependents[module]:
                indegree[dependent] -= 1
                if indegree[dependent] == 0:
                    heappush(ready, (module_indexes[dependent], dependent))

        if len(ordered) != len(runnable):
            cyclic_modules = sorted(
                (module.__name__ for module, count in indegree.items() if count > 0)
            )
            self.log.warning(
                "Circular module dependency detected involving: %s. "
                "No modules will be run.",
                ", ".join(cyclic_modules),
            )
            return None

        return ordered

    def run(self) -> None:
        ordered_modules = self._ordered_modules()
        if ordered_modules is None:
            return

        self._log_loaded_modules(ordered_modules)

        try:
            self.init()
        except NotImplementedError:
            pass

        executed_by_type: dict[type[MVTModule], MVTModule] = {}
        for module in ordered_modules:

            module_logger = get_module_logger(module)

            m = module(
                target_path=self.target_path,
                results_path=self.results_path,
                module_options=self.module_options,
                log=module_logger,
            )
            # Dependencies are keyed by the module class they declare, even
            # when it was replaced, so that a module asking for the results
            # of a replaced module receives those of its replacement.
            m.dependency_modules = {
                dependency: executed_by_type[resolved]
                for dependency, resolved in self._module_dependencies(module)
            }

            if self.iocs.total_ioc_count:
                m.indicators = self.iocs
                m.indicators.log = m.log

            if self.serial:
                m.serial = self.serial

            try:
                self.module_init(m)
            except NotImplementedError:
                pass

            try:
                run_module(m)
            except EncryptedBackupError:
                self.log.critical(
                    "The backup appears to be encrypted. "
                    "Please decrypt it first using `mvt-ios decrypt-backup`."
                )
                return

            self.executed.append(m)
            executed_by_type[module] = m
            self.timeline.extend(m.timeline)
            self.url_results.extend(m.url_results)
            self.alertstore.extend(m.alertstore.alerts)

        try:
            self.finish()
        except NotImplementedError:
            pass

        # We only store the timeline from the parent/main command
        if self.sub_command:
            return

        self._store_timeline()
        self._store_alerts_timeline()
        self._store_alerts()
        self._store_urls()
        self._store_info()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import hashlib
import logging
import os
import plistlib
from datetime import datetime, timezone
from typing import Any, Dict, Optional

from mvt.common.module import DatabaseNotFoundError
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_datetime_to_iso

from ..base import IOSExtraction

APPLICATIONS_DB_PATH = [
    "private/var/containers/Bundle/Application/*/iTunesMetadata.plist"
]
KNOWN_APP_INSTALLERS = [
    "com.apple.AppStore",
    "com.apple.AppStore.ProductPageExtension",
    "com.apple.dmd",
    "dmd",
]


class Applications(IOSExtraction):
    """Extract information from accounts installed on the phone."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        if "isodate" in record:
            return {
                "timestamp": record["isodate"],
                "module": self.__class__.__name__,
                "event": "app_installed",
                "data": f"App {record.get('name', '')} version {record.get('bundleShortVersionString', '')} from {record.get('artistName', '')} installed from {record.get('sourceApp', '')}",
            }
        return []

    def check_indicators(self) -> None:
        for result in self.results:
            if self.indicators:
                if "softwareVersionBundleId" not in result:
                    self.alertstore.medium(
                        "Suspicious application identified without softwareVersionBundleId",
                        "",
                        result,
                    )
                    continue

                ioc_match = self.indicators.check_process(
                    result["softwareVersionBundleId"]
                )
                if ioc_match:
                    self.alertstore.critical(
                        f"Malicious application {result['softwareVersionBundleId']} identified",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )
                    continue

                ioc_match = self.indicators.check_app_id(
                    result["softwareVersionBundleId"]
                )
                if ioc_match:
                    self.alertstore.critical(
                        f"Malicious application {result['softwareVersionBundleId']} identified",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )
                    continue

            # Some apps installed from apple store with sourceApp "com.apple.AppStore.ProductPageExtension"
            if (
                result.get("sourceApp", "com.apple.AppStore")
                not in KNOWN_APP_INSTALLERS
            ):
                self.alertstore.medium(
                    f"Suspicious app not installed from the App Store or MDM: {result['softwareVersionBundleId']}",
                    "",
                    result,
                )

    def _parse_itunes_timestamp(self, entry: Dict[str, Any]) -> None:
        """
        Parse the iTunes metadata info
        """
        if entry.get("com.apple.iTunesStore.downloadInfo", {}).get(
            "purchaseDate", None
        ):
            timestamp = datetime.strptime(
                entry["com.apple.iTunesStore.downloadInfo"]["purchaseDate"],
                "%Y-%m-%dT%H:%M:%SZ",
            )
            timestamp_utc = timestamp.astimezone(timezone.utc)
            entry["isodate"] = convert_datetime_to_iso(timestamp_utc)

    def _parse_itunes_metadata(self, plist_path: str) -> None:
        """
        Parse iTunesMetadata.plist file from an application in fs dump
        """
        with open(plist_path, "rb") as f:
            entry = plistlib.load(f)

        entry["file_path"] = plist_path
        self._parse_itunes_timestamp(entry)
        self.results.append(entry)

    def _parse_info_plist(self, plist_path: str) -> None:
        """
        Parse Info.plist file from backup
        """
        with open(plist_path, "rb") as f:
            data = plistlib.load(f)

        for app in data.get("Applications", {}):
            app_data = data["Applications"][app]
            entry = {"name": app}
            metadata = plistlib.loads(app_data["iTunesMetadata"])
            entry.update(metadata)

            self._parse_itunes_timestamp(entry)

            if "PlaceholderIcon" in app_data:
                sha256_hash = hashlib.sha256()
                sha256_hash.update(app_data["PlaceholderIcon"])
                entry["icon_sha256"] = sha256_hash.hexdigest()

            self.results.append(entry)

    def run(self) -> None:
        if self.is_backup:
            if not self.target_path:
                return
            plist_path = os.path.join(self.target_path, "Info.plist")
            if not os.path.isfile(plist_path):
                raise DatabaseNotFoundError("Impossible to find Info.plist file")
            self._parse_info_plist(plist_path)
        elif self.is_fs_dump:
            for file_path in self._get_fs_files_from_patterns(APPLICATIONS_DB_PATH):
                self._parse_itunes_metadata(file_path)

        self.log.info("Extracted a total of %d applications", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import datetime
import json
import logging
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_datetime_to_iso

from ..base import IOSExtraction

IOS_ANALYTICS_JOURNAL_PATHS = [
    "private/var/db/analyticsd/Analytics-Journal-*.ips",
]


class IOSVersionHistory(IOSExtraction):
    """This module extracts iOS update history from Analytics Journal log files."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.results: list = [] if results is None else results

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "ios_version",
            "data": f"Recorded iOS version {record['os_version']}",
        }

    def run(self) -> None:
        for found_path in self._get_fs_files_from_patterns(IOS_ANALYTICS_JOURNAL_PATHS):
            with open(found_path, "r", encoding="utf-8") as analytics_log:
                log_line = json.loads(analytics_log.readline().strip())

                timestamp = datetime.datetime.strptime(
                    log_line["timestamp"], "%Y-%m-%d %H:%M:%S.%f %z"
                )
                timestamp_utc = timestamp.astimezone(datetime.timezone.utc)
                self.results.append(
                    {
                        "isodate": convert_datetime_to_iso(timestamp_utc),
                        "os_version": log_line["os_version"],
                    }
                )

        self.results = sorted(self.results, key=lambda entry: entry["isodate"])



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import json
import logging
from typing import Optional

from .base import AndroidQFModule


class RootBinaries(AndroidQFModule):
    """This module analyzes root_binaries.json for root binaries found by androidqf."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[list] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: dict) -> dict:
        return {
            "timestamp": record.get("timestamp"),
            "module": self.__class__.__name__,
            "event": "root_binary_found",
            "data": f"Root binary found: {record['path']} (binary: {record['binary_name']})",
        }

    def check_indicators(self) -> None:
        """Check for indicators of device rooting."""
        if not self.results:
            return

        # All found root binaries are considered indicators of rooting
        for result in self.results:
            self.alertstore.high(
                f'Found root binary "{result["binary_name"]}" at path "{result["path"]}"',
                "",
                result,
            )

        if self.results:
            self.log.warning(
                "Device shows signs of rooting with %d root binaries found",
                len(self.results),
            )

    def run(self) -> None:
        """Run the root binaries analysis."""
        root_binaries_files = self._get_files_by_pattern("*/root_binaries.json")

        if not root_binaries_files:
            self.log.info("No root_binaries.json file found")
            return

        rawdata = self._get_file_content(root_binaries_files[0]).decode(
            "utf-8", errors="ignore"
        )

        try:
            root_binary_paths = json.loads(rawdata)
        except json.JSONDecodeError as e:
            self.log.error("Failed to parse root_binaries.json: %s", e)
            return

        if not isinstance(root_binary_paths, list):
            self.log.error("Expected root_binaries.json to contain a list of paths")
            return

        # Known root binary names that might be found and their descriptions
        # This maps the binary name to a human-readable description
        known_root_binaries = {
            "su": "SuperUser binary",
            "busybox": "BusyBox utilities",
            "supersu": "SuperSU root management",
            "Superuser.apk": "Superuser app",
            "KingoUser.apk": "KingRoot app",
            "SuperSu.apk": "SuperSU app",
            "magisk": "Magisk root framework",
            "magiskhide": "Magisk hide utility",
            "magiskinit": "Magisk init binary",
            "magiskpolicy": "Magisk policy binary",
        }

        for path in root_binary_paths:
            if not path or not isinstance(path, str):
                continue

            # Extract binary name from path
            binary_name = path.split("/")[-1].lower()

            # Check if this matches a known root binary by exact name match
            description = "Unknown root binary"
            for known_binary in known_root_binaries:
                if binary_name == known_binary.lower():
                    description = known_root_binaries[known_binary]
                    break

            result = {
                "path": path.strip(),
                "binary_name": binary_name,
                "description": description,
            }

            self.results.append(result)

        self.log.info("Found %d root binaries", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2026 The MVT Authors.

from mvt.android.artifacts.processes import Processes as ProcessesArtifact

from .base import BugReportModule


class Processes(ProcessesArtifact, BugReportModule):
    """Extract the process and thread table from dumpstate."""

    def run(self) -> None:
        data = self._get_dumpstate_file()
        if not data:
            self.log.error("Unable to find dumpstate file")
            return
        section = self.extract_command_section(
            data.decode("utf-8", errors="replace"),
            "------ PROCESSES AND THREADS",
        )
        self.parse(section)
        self.log.info("Identified %d running process threads", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import binascii
import glob
import logging
import multiprocessing
import os
import os.path
import shutil
import sqlite3
from pathlib import Path
from typing import Optional

from iOSbackup import iOSbackup

log = logging.getLogger(__name__)


class DecryptBackup:
    """This class provides functions to decrypt an encrypted iTunes backup
    using either a password or a key file.


    """

    def __init__(self, backup_path: str, dest_path: Optional[str] = None) -> None:
        """Decrypts an encrypted iOS backup.
        :param backup_path: Path to the encrypted backup folder
        :param dest_path: Path to the folder where to store the decrypted backup
        """
        self.backup_path = os.path.abspath(backup_path)
        self.dest_path = dest_path
        self._backup = None
        self._decryption_key = None

    def can_process(self) -> bool:
        return self._backup is not None

    @staticmethod
    def is_encrypted(backup_path: str) -> bool:
        """Query Manifest.db file to see if it's encrypted or not.

        :param backup_path: Path to the backup to decrypt

        """
        conn = sqlite3.connect(os.path.join(backup_path, "Manifest.db"))
        try:
            cur = conn.cursor()
            cur.execute("SELECT fileID FROM Files LIMIT 1;")
        except sqlite3.DatabaseError:
            return True
        else:
            log.critical("The backup does not seem encrypted!")
            return False
        finally:
            conn.close()

    def _process_file(
        self, relative_path: str, domain: str, item, file_id: str, item_folder: str
    ) -> None:
        assert self._backup is not None
        self._backup.getFileDecryptedCopy(
            manifestEntry=item, targetName=file_id, targetFolder=item_folder
        )
        log.info(
            "Decrypted file %s [%s] to %s/%s",
            relative_path,
            domain,
            item_folder,
            file_id,
        )

    def process_backup(self) -> None:
        assert self._backup is not None
        assert self.dest_path is not None

        if not os.path.exists(self.dest_path):
            os.makedirs(self.dest_path)

        manifest_path = os.path.join(self.dest_path, "Manifest.db")
        # We extract a decrypted Manifest.db.
        self._backup.getManifestDB()
        # We store it to the destination folder.
        shutil.copy(self._backup.manifestDB, manifest_path)

        pool = multiprocessing.Pool(multiprocessing.cpu_count())

        for item in self._backup.getBackupFilesList():
            try:
                file_id = item["backupFile"]
                relative_path = item["relativePath"]
                domain = item["domain"]

                # This may be a partial backup. Skip files from the manifest
                # which do not exist locally.
                source_file_path = os.path.join(self.backup_path, file_id[0:2], file_id)
                if not Path(source_file_path).resolve().is_relative_to(Path(self.backup_path).resolve()):
                    log.warning("Skipping unsafe file_id: %r", file_id)
                    continue
                if not os.path.exists(source_file_path):
                    log.debug(
                        "Skipping file %s. File not found in encrypted backup directory.",
                        source_file_path,
                    )
                    continue

                item_folder = os.path.join(self.dest_path, file_id[0:2])  # type: ignore[arg-type]
                if not Path(os.path.join(item_folder, file_id)).resolve().is_relative_to(Path(self.dest_path).resolve()):
                    log.warning("Skipping unsafe file_id: %r", file_id)
                    continue
                if not os.path.exists(item_folder):
                    os.makedirs(item_folder)

                # iOSBackup getFileDecryptedCopy() claims to read a "file"
                # parameter but the code actually is reading the "manifest" key.
                # Add manifest plist to both keys to handle this.
                item["manifest"] = item["file"]

                pool.apply_async(
                    self._process_file,
                    args=(relative_path, domain, item, file_id, item_folder),
                )
            except Exception as exc:
                log.error("Failed to decrypt file %s: %s", relative_path, exc)

        pool.close()
        pool.join()

        # Copying over the root plist files as well.
        for file_name in os.listdir(self.backup_path):
            if file_name.endswith(".plist"):
                log.info("Copied plist file %s to %s", file_name, self.dest_path)
                shutil.copy(os.path.join(self.backup_path, file_name), self.dest_path)

    def decrypt_with_password(self, password: str) -> None:
        """Decrypts an encrypted iOS backup.

        :param password: Password to use to decrypt the original backup

        """
        log.info("Decrypting iOS backup at path %s with password", self.backup_path)

        if not os.path.exists(os.path.join(self.backup_path, "Manifest.plist")):
            possible = glob.glob(os.path.join(self.backup_path, "*", "Manifest.plist"))

            if len(possible) == 1:
                newpath = os.path.dirname(possible[0])
                log.warning(
                    "No Manifest.plist in %s, using %s instead.",
                    self.backup_path,
                    newpath,
                )
                self.backup_path = newpath
            elif len(possible) > 1:
                log.critical(
                    "No Manifest.plist in %s, and %d Manifest.plist files in subdirs. "
                    "Please choose one!",
                    self.backup_path,
                    len(possible),
                )
                return

        # Before proceeding, we check whether the backup is indeed encrypted.
        if not self.is_encrypted(self.backup_path):
            return

        try:
            self._backup = iOSbackup(
                udid=os.path.basename(self.backup_path),
                cleartextpassword=password,
                backuproot=os.path.dirname(self.backup_path),
            )
        except Exception as exc:
            if (
                isinstance(exc, KeyError)
                and len(exc.args) > 0
                and exc.args[0] == b"KEY"
            ):
                log.critical("Failed to decrypt backup. Password is probably wrong.")
            elif (
                isinstance(exc, FileNotFoundError)
                and os.path.basename(exc.filename) == "Manifest.plist"
            ):
                log.critical(
                    "Failed to find a valid backup at %s. "
                    "Did you point to the right backup path?",
                    self.backup_path,
                )
            else:
                log.exception(exc)
                log.critical(
                    "Failed to decrypt backup. Did you provide the correct password? "
                    "Did you point to the right backup path?"
                )

    def decrypt_with_key_file(self, key_file: str) -> None:
        """Decrypts an encrypted iOS backup using a key file.

        :param key_file: File to read the key bytes to decrypt the backup

        """
        log.info(
            "Decrypting iOS backup at path %s with key file %s",
            self.backup_path,
            key_file,
        )

        # Before proceeding, we check whether the backup is indeed encrypted.
        if not self.is_encrypted(self.backup_path):
            return

        with open(key_file, "rb") as handle:
            key_bytes = handle.read()

        # Key should be 64 hex encoded characters (32 raw bytes)
        if len(key_bytes) != 64:
            log.critical(
                "Invalid key from key file. Did you provide the correct key file?"
            )
            return

        try:
            key_bytes_raw = binascii.unhexlify(key_bytes)
            self._backup = iOSbackup(
                udid=os.path.basename(self.backup_path),
                derivedkey=key_bytes_raw,
                backuproot=os.path.dirname(self.backup_path),
            )
        except Exception as exc:
            log.exception(exc)
            log.critical(
                "Failed to decrypt backup. Did you provide the correct key file?"
            )

    def get_key(self) -> None:
        """Retrieve and prints the encryption key."""
        if not self._backup:
            return

        self._decryption_key = self._backup.getDecryptionKey()
        log.info(
            'Derived decryption key for backup at path %s is: "%s"',
            self.backup_path,
            self._decryption_key,
        )

    def write_key(self, key_path: str) -> None:
        """Save extracted key to file.

        :param key_path: Path to the file where to write the derived decryption
                         key.

        """
        if not self._decryption_key:
            return

        try:
            with open(key_path, "w", encoding="utf-8") as handle:
                handle.write(self._decryption_key)
        except Exception as exc:
            log.exception(exc)
            log.critical("Failed to write key to file: %s", key_path)
            return
        else:
            log.info(
                "Wrote decryption key to file: %s. This file is "
                "equivalent to a plaintext password. Keep it safe!",
                key_path,
            )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
import sqlite3
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)

from ..base import IOSExtraction


class CacheFiles(IOSExtraction):
    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        records = []
        for item in self.results[record]:
            records.append(
                {
                    "timestamp": item["isodate"],
                    "module": self.__class__.__name__,
                    "event": "cache_response",
                    "data": f"{record} recorded visit to URL {item['url']}",
                }
            )

        return records

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for key, values in self.results.items():
            for value in values:
                ioc_match = self.indicators.check_url(value["url"])
                if ioc_match:
                    value["cache_file"] = key
                    self.alertstore.critical(
                        ioc_match.message,
                        value.get("isodate", ""),
                        value,
                        matched_indicator=ioc_match.ioc,
                    )

    def _process_cache_file(self, file_path):
        self.log.info("Processing cache file at path: %s", file_path)

        conn = self._open_sqlite_db(file_path)
        cur = conn.cursor()

        try:
            try:
                cur.execute("SELECT * FROM cfurl_cache_response;")
            except sqlite3.OperationalError:
                return

            key_name = os.path.relpath(file_path, self.target_path)
            if key_name not in self.results:
                self.results[key_name] = []

            for row in cur:
                self.results[key_name].append(
                    {
                        "entry_id": row[0],
                        "version": row[1],
                        "hash_value": row[2],
                        "storage_policy": row[3],
                        "url": row[4],
                        "isodate": row[5],
                    }
                )
        finally:
            cur.close()
            conn.close()

    def run(self) -> None:
        self.results: dict = {}
        if not self.target_path:
            self.log.error("No filesystem dump path provided")
            return

        for root, _, files in os.walk(self.target_path):
            for file_name in files:
                if file_name != "Cache.db":
                    continue

                file_path = os.path.join(root, file_name)
                self._process_cache_file(file_path)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import base64
import logging
import plistlib
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

LOCATIOND_BACKUP_IDS = [
    "a690d7769cce8904ca2b67320b107c8fe5f79412",
]
LOCATIOND_ROOT_PATHS = [
    "private/var/mobile/Library/Caches/locationd/clients.plist",
    "private/var/root/Library/Caches/locationd/clients.plist",
]


class LocationdClients(IOSExtraction):
    """Extract information from apps who used geolocation."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

        self.timestamps = [
            "ConsumptionPeriodBegin",
            "ReceivingLocationInformationTimeStopped",
            "VisitTimeStopped",
            "LocationTimeStopped",
            "BackgroundLocationTimeStopped",
            "SignificantTimeStopped",
            "NonPersistentSignificantTimeStopped",
            "FenceTimeStopped",
            "BeaconRegionTimeStopped",
        ]

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        records = []
        for timestamp in self.timestamps:
            if timestamp in record.keys():
                records.append(
                    {
                        "timestamp": record[timestamp],
                        "module": self.__class__.__name__,
                        "event": timestamp,
                        "data": f"{timestamp} from {record['package']}",
                    }
                )

        return records

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            parts = result["package"].split("/")
            proc_name = parts[len(parts) - 1]

            ioc_match = self.indicators.check_process(proc_name)
            if ioc_match:
                self.alertstore.high(
                    f"Found a suspicious process name in LocationD entry {result['package']}",
                    "",
                    result,
                    matched_indicator=ioc_match.ioc,
                )
                continue

            if "BundleId" in result:
                ioc_match = self.indicators.check_process(result["BundleId"])
                if ioc_match:
                    self.alertstore.high(
                        f"Found a suspicious process name in LocationD entry {result['package']}",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )

            if "BundlePath" in result:
                ioc_match = self.indicators.check_file_path(result["BundlePath"])
                if ioc_match:
                    self.alertstore.high(
                        f"Found a known malicious domain in LocationD entry {result['package']}",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )
                    continue

            if "Executable" in result:
                ioc_match = self.indicators.check_file_path(result["Executable"])
                if ioc_match:
                    self.alertstore.high(
                        f"Found a suspicious file path in LocationD entry {result['Executable']}",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )
                    continue

            if "Registered" in result:
                # Sometimes registered is a bool
                if isinstance(result["Registered"], bool):
                    continue

                ioc_match = self.indicators.check_file_path(result["Registered"])
                if ioc_match:
                    self.alertstore.high(
                        f"Found a suspicious file path in LocationD entry {result['Registered']}",
                        "",
                        result,
                        matched_indicator=ioc_match.ioc,
                    )
                    continue

    def _extract_locationd_entries(self, file_path):
        with open(file_path, "rb") as handle:
            file_plist = plistlib.load(handle)

        for key, _ in file_plist.items():
            # Some migration information are int and not dicts
            if not isinstance(file_plist[key], dict):
                continue
            result = file_plist[key]
            result["package"] = key.rstrip(":")
            for timestamp in self.timestamps:
                if timestamp in result.keys():
                    result[timestamp] = convert_mactime_to_iso(result[timestamp])

            if "ClientStorageToken" in result:
                result["ClientStorageToken"] = base64.b64encode(
                    result["ClientStorageToken"]
                )

            self.results.append(result)

    def run(self) -> None:
        if self.is_backup:
            self._find_ios_database(backup_ids=LOCATIOND_BACKUP_IDS)
            self.log.info("Found Locationd Clients plist at path: %s", self.file_path)
            self._extract_locationd_entries(self.file_path)
        elif self.is_fs_dump:
            for locationd_path in self._get_fs_files_from_patterns(
                LOCATIOND_ROOT_PATHS
            ):
                self.file_path = locationd_path
                self.log.info(
                    "Found Locationd Clients plist at path: %s", self.file_path
                )
                self._extract_locationd_entries(self.file_path)

        self.log.info(
            "Extracted a total of %d Locationd Clients entries", len(self.results)
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2026 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import sqlite3
from typing import Optional

from mvt.common.module import DatabaseNotFoundError
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

WHATSAPP_CONTACTS_BACKUP_IDS = [
    # SHA-1 of "AppDomainGroup-group.net.whatsapp.WhatsApp.shared-ContactsV2.sqlite"
    "b8548dc30aa1030df0ce18ef08b882cf7ab5212f",
]
WHATSAPP_CONTACTS_ROOT_PATHS = [
    "private/var/mobile/Containers/Shared/AppGroup/*/ContactsV2.sqlite",
]

# WhatsApp's standard disappearing-messages timer values, in seconds.
DISAPPEARING_DURATION_LABELS = {
    86400: "24 hours",
    604800: "7 days",
    1209600: "14 days",
    2592000: "30 days",
    7776000: "90 days",
}

# Output field -> candidate columns in ZWAADDRESSBOOKCONTACT, in order of
# preference. WhatsApp renames columns across versions, so the query is built
# from the columns actually present in the database.
COLUMN_CANDIDATES = {
    "whatsapp_id": ["ZWHATSAPPID"],
    "lid": ["ZLID"],
    "phone_number": ["ZPHONENUMBER"],
    "localized_phone_number": ["ZLOCALIZEDPHONENUMBER"],
    "full_name": ["ZFULLNAME"],
    "given_name": ["ZGIVENNAME"],
    "last_name": ["ZLASTNAME"],
    "user_name": ["ZUSERNAME"],
    "business_name": ["ZBUSINESSNAME"],
    "about_text": ["ZABOUTTEXT"],
    "about_emoji": ["ZABOUTEMOJI"],
    "notes": ["ZNOTES"],
    "disappearing_mode_duration": ["ZDISAPPEARINGMODEDURATION"],
    "disappearing_mode_timestamp": ["ZDISAPPEARINGMODETIMESTAMP"],
    "about_timestamp": ["ZABOUTTIMESTAMP"],
    "about_expiration_timestamp": ["ZABOUTEXPIRATIONTIMESTAMP"],
    "last_updated": ["ZLASTUPDATED"],
    "phone_status": ["ZPHONESTATUS", "ZPHONENUMBERSTATUS"],
    "sync_policy": ["ZSYNCPOLICY"],
}

STRING_FIELDS = [
    "whatsapp_id",
    "lid",
    "phone_number",
    "localized_phone_number",
    "full_name",
    "given_name",
    "last_name",
    "user_name",
    "business_name",
    "about_text",
    "about_emoji",
    "notes",
]

DATE_FIELDS = [
    "disappearing_mode_timestamp",
    "about_timestamp",
    "about_expiration_timestamp",
    "last_updated",
]


def _decode_string(value) -> Optional[str]:
    # CoreData stores string attributes as UTF-8 blobs in some WhatsApp
    # versions, so values can arrive as either bytes or str.
    if value is None:
        return None
    if isinstance(value, bytes):
        return value.decode("utf-8", "replace")
    return str(value)


def _label_duration(duration) -> str:
    if not duration:
        return "off"
    return DISAPPEARING_DURATION_LABELS.get(
        int(duration), f"{int(duration)} seconds"
    )


def _describe_contact(record: ModuleAtomicResult) -> str:
    contact = (
        record.get("whatsapp_id")
        or record.get("lid")
        or record.get("phone_number")
        or "unknown"
    )
    full_name = record.get("full_name")
    if full_name:
        contact = f"{contact} ({full_name})"
    return contact


class WhatsappContacts(IOSExtraction):
    """This module extracts WhatsApp contact records and per-contact
    disappearing-messages settings from ContactsV2.sqlite.

    ChatStorage.sqlite does not record the disappearing-messages state of 1:1
    chats: the authoritative timer is stored on each contact record in this
    database, alongside the mapping between a contact's LID and phone number
    identifiers.
    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        records = []
        contact = _describe_contact(record)

        if record.get("disappearing_mode_timestamp"):
            records.append(
                {
                    "timestamp": record["disappearing_mode_timestamp"],
                    "module": self.__class__.__name__,
                    "event": "disappearing_mode_set",
                    "data": (
                        f"WhatsApp disappearing messages timer set to "
                        f"'{record.get('disappearing_mode_label')}' "
                        f"for {contact}"
                    ),
                }
            )

        if record.get("about_timestamp"):
            data = f"WhatsApp about text of {contact} changed"
            about_text = record.get("about_text")
            if about_text:
                data += f' to "{about_text}"'
            records.append(
                {
                    "timestamp": record["about_timestamp"],
                    "module": self.__class__.__name__,
                    "event": "about_changed",
                    "data": data,
                }
            )

        if record.get("about_expiration_timestamp"):
            records.append(
                {
                    "timestamp": record["about_expiration_timestamp"],
                    "module": self.__class__.__name__,
                    "event": "about_expiration",
                    "data": (
                        f"WhatsApp about text of {contact} scheduled "
                        f"to expire"
                    ),
                }
            )

        if record.get("last_updated"):
            records.append(
                {
                    "timestamp": record["last_updated"],
                    "module": self.__class__.__name__,
                    "event": "contact_last_updated",
                    "data": f"WhatsApp contact record for {contact} updated",
                }
            )

        return records

    def run(self) -> None:
        try:
            self._find_ios_database(
                backup_ids=WHATSAPP_CONTACTS_BACKUP_IDS,
                root_paths=WHATSAPP_CONTACTS_ROOT_PATHS,
            )
        except DatabaseNotFoundError:
            self.log.warning(
                "Unable to find the WhatsApp ContactsV2.sqlite database in "
                "this backup or filesystem dump. WhatsApp disappearing "
                "messages settings and contact records cannot be extracted. "
                "This database is often missing from incremental backups."
            )
            return

        self.log.info(
            "Found WhatsApp contacts database at path: %s", self.file_path
        )

        assert self.file_path is not None
        conn = self._open_sqlite_db(self.file_path)
        cur = conn.cursor()
        try:
            try:
                cur.execute("PRAGMA table_info(ZWAADDRESSBOOKCONTACT)")
                available_columns = {row[1] for row in cur.fetchall()}
            except sqlite3.DatabaseError as exc:
                self.log.error(
                    "Unable to read the ZWAADDRESSBOOKCONTACT table schema: %s",
                    exc,
                )
                return

            if not available_columns:
                self.log.warning(
                    "The WhatsApp contacts database does not contain a "
                    "ZWAADDRESSBOOKCONTACT table"
                )
                return

            selected = {}
            for field, candidates in COLUMN_CANDIDATES.items():
                for candidate in candidates:
                    if candidate in available_columns:
                        selected[field] = candidate
                        break

            # A record with no duration column is "unknown", not "off": the
            # timer state cannot be determined from this database version.
            has_duration = "disappearing_mode_duration" in selected
            if not has_duration:
                self.log.warning(
                    "The ZDISAPPEARINGMODEDURATION column is not present in "
                    "this WhatsApp contacts database: disappearing messages "
                    "state is unknown"
                )

            columns = ["Z_PK"] + list(selected.values())
            cur.execute(
                f"SELECT {', '.join(columns)} FROM ZWAADDRESSBOOKCONTACT;"
            )
            fields = ["row_pk"] + list(selected.keys())

            for row in cur:
                record = dict(zip(fields, row))

                for field in STRING_FIELDS:
                    if field in record:
                        record[field] = _decode_string(record[field])
                    else:
                        record[field] = None

                for field in DATE_FIELDS:
                    if record.get(field) is not None:
                        record[field] = (
                            convert_mactime_to_iso(record[field]) or None
                        )
                    else:
                        record[field] = None

                duration = record.get("disappearing_mode_duration")
                if has_duration:
                    record["disappearing_mode_is_on"] = bool(duration)
                    record["disappearing_mode_label"] = _label_duration(
                        duration
                    )
                else:
                    record["disappearing_mode_duration"] = None
                    record["disappearing_mode_is_on"] = None
                    record["disappearing_mode_label"] = None

                record.setdefault("phone_status", None)
                record.setdefault("sync_policy", None)

                self.results.append(record)
        finally:
            cur.close()
            conn.close()

        total_ephemeral = sum(
            1 for record in self.results if record["disappearing_mode_is_on"]
        )
        self.log.info(
            "Extracted a total of %d WhatsApp contacts (%d with disappearing "
            "messages enabled)",
            len(self.results),
            total_ephemeral,
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Optional

from mvt.common.command import Command
from mvt.common.indicators import Indicators
from mvt.common.module import MVTModule

from .modules.fs import FS_MODULES
from .modules.mixed import MIXED_MODULES

log = logging.getLogger(__name__)


class CmdIOSCheckFS(Command):
    def __init__(
        self,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        ioc_files: Optional[list] = None,
        iocs: Optional[Indicators] = None,
        module_name: Optional[str] = None,
        serial: Optional[str] = None,
        module_options: Optional[dict] = None,
        hashes: bool = False,
        sub_command: bool = False,
        disable_version_check: bool = False,
        disable_indicator_check: bool = False,
        custom_modules: Optional[list[type[MVTModule]]] = None,
    ) -> None:
        super().__init__(
            target_path=target_path,
            results_path=results_path,
            ioc_files=ioc_files,
            iocs=iocs,
            module_name=module_name,
            module_options=module_options,
            hashes=hashes,
            sub_command=sub_command,
            log=log,
            disable_version_check=disable_version_check,
            disable_indicator_check=disable_indicator_check,
            custom_modules=custom_modules,
        )

        self.platform = "ios"
        self.name = "check-fs"
        self.modules = FS_MODULES + MIXED_MODULES

    def module_init(self, module):
        module.is_fs_dump = True



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
from pathlib import Path
from typing import List, Optional
from zipfile import ZipFile

from mvt.android.modules.bugreport.base import BugReportModule
from mvt.common.command import Command
from mvt.common.indicators import Indicators
from mvt.common.module import MVTModule

from .modules.bugreport import BUGREPORT_MODULES

log = logging.getLogger(__name__)


class CmdAndroidCheckBugreport(Command):
    def __init__(
        self,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        ioc_files: Optional[list] = None,
        iocs: Optional[Indicators] = None,
        module_name: Optional[str] = None,
        serial: Optional[str] = None,
        module_options: Optional[dict] = None,
        hashes: Optional[bool] = False,
        sub_command: Optional[bool] = False,
        disable_version_check: bool = False,
        disable_indicator_check: bool = False,
        custom_modules: Optional[list[type[MVTModule]]] = None,
    ) -> None:
        super().__init__(
            target_path=target_path,
            results_path=results_path,
            ioc_files=ioc_files,
            iocs=iocs,
            module_name=module_name,
            serial=serial,
            module_options=module_options,
            hashes=hashes,
            sub_command=sub_command,
            log=log,
            disable_version_check=disable_version_check,
            disable_indicator_check=disable_indicator_check,
            custom_modules=custom_modules,
        )

        self.platform = "android"
        self.name = "check-bugreport"
        self.modules = BUGREPORT_MODULES

        self.__format: str = ""
        self.__zip: Optional[ZipFile] = None
        self.__files: List[str] = []

    def from_dir(self, dir_path: str) -> None:
        """This method is used to initialize the bug report analysis from an
        uncompressed directory.
        """
        self.__format = "dir"
        self.target_path = dir_path
        parent_path = Path(dir_path).absolute().as_posix()
        for root, _, subfiles in os.walk(os.path.abspath(dir_path)):
            for file_name in subfiles:
                file_path = os.path.relpath(os.path.join(root, file_name), parent_path)
                self.__files.append(file_path)

    def from_zip(self, bugreport_zip: ZipFile) -> None:
        """This method is used to initialize the bug report analysis from a
        compressed archive.
        """
        # NOTE: This will be invoked either by the CLI directly,or by the
        # check-androidqf command. We need this because we want to support
        # check-androidqf to analyse compressed archives itself too.
        # So, we'll need to extract bugreport.zip from a 'androidqf.zip', and
        # since nothing is written on disk, we need to be able to pass this
        # command a ZipFile instance in memory.

        self.__format = "zip"
        self.__zip = bugreport_zip
        for file_name in self.__zip.namelist():
            self.__files.append(file_name)

    def init(self) -> None:
        if not self.target_path:
            return

        if os.path.isfile(self.target_path):
            self.from_zip(ZipFile(self.target_path))
        elif os.path.isdir(self.target_path):
            self.from_dir(self.target_path)

    def module_init(self, module: BugReportModule) -> None:  # type: ignore[override]
        if self.__format == "zip":
            module.from_zip(self.__zip, self.__files)
        else:
            if not self.target_path:
                raise ValueError("target_path is not set")
            module.from_dir(self.target_path, self.__files)

    def finish(self) -> None:
        if self.__zip:
            self.__zip.close()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import datetime
from typing import Optional
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from mvt.common.utils import convert_datetime_to_iso
from .base import BugReportModule
from mvt.common.module_types import ModuleResults
from mvt.android.artifacts.file_timestamps import FileTimestampsArtifact
from mvt.android.artifacts.getprop import GetProp


class BugReportTimestamps(FileTimestampsArtifact, BugReportModule):
    """This module extracts records from battery daily updates."""

    slug = "bugreport_timestamps"

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def run(self) -> None:
        filesystem_files = self._get_files_by_pattern("FS/*")

        timezone_name = None
        dumpstate = self._get_dumpstate_file()
        if dumpstate:
            section = self.extract_command_section(
                dumpstate.decode("utf-8", errors="replace"),
                "------ SYSTEM PROPERTIES",
            )
            properties = GetProp()
            properties.parse(section)
            timezone_name = properties.get_device_timezone()
        timezone = None
        if timezone_name:
            try:
                timezone = ZoneInfo(timezone_name)
            except ZoneInfoNotFoundError:
                self.log.warning("Unknown device timezone %s", timezone_name)

        self.results = []
        for file in filesystem_files:
            # Only the modification time is available in the zip file metadata.
            # The timezone is the local timezone of the machine the phone.
            modification_time = self._get_file_modification_time(file)
            utc_time = None
            if timezone is not None:
                utc_time = convert_datetime_to_iso(
                    modification_time.replace(tzinfo=timezone).astimezone(
                        datetime.timezone.utc
                    )
                )
            self.results.append(
                {
                    "path": file,
                    "modified_time": convert_datetime_to_iso(modification_time),
                    "modified_time_utc": utc_time,
                    "timezone": timezone_name,
                    "timestamp_source": (
                        "zip_metadata" if self.zip_archive else "filesystem_metadata"
                    ),
                }
            )

        self.log.info(
            "Extracted a total of %d filesystem timestamps from bugreport.",
            len(self.results),
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import datetime
from typing import List, Optional

import pydantic
import betterproto2
from dateutil import parser

from mvt.android.parsers.proto.tombstone import Tombstone
from mvt.common.module_types import ModuleAtomicResult, ModuleSerializedResult
from mvt.common.utils import convert_datetime_to_iso

from .artifact import AndroidArtifact

TOMBSTONE_DELIMITER = "*** *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***"

# Map the legacy crash file keys to the new format.
TOMBSTONE_TEXT_KEY_MAPPINGS = {
    "Build fingerprint": "build_fingerprint",
    "Revision": "revision",
    "ABI": "arch",
    "Timestamp": "timestamp",
    "Process uptime": "process_uptime",
    "Cmdline": "command_line",
    "pid": "pid",
    "tid": "tid",
    "name": "process_name",
    "binary_path": "binary_path",
    "uid": "uid",
    "signal": "signal_info",
    "code": "code",
    "Cause": "cause",
    "Abort message": "abort_message",
}


class SignalInfo(pydantic.BaseModel):
    code: int
    code_name: str
    name: str
    number: Optional[int] = None


class TombstoneCrashResult(pydantic.BaseModel):
    """
    MVT Result model for a tombstone crash result.

    Needed for validation and serialization, and consistency between text and protobuf tombstones.
    """

    file_name: str
    file_timestamp: str  # We store the timestamp as a string to avoid timezone issues
    build_fingerprint: str
    revision: str
    arch: Optional[str] = None
    timestamp: str  # We store the timestamp as a string to avoid timezone issues
    process_uptime: Optional[int] = None
    command_line: Optional[List[str]] = None
    pid: int
    tid: int
    process_name: Optional[str] = None
    binary_path: Optional[str] = None
    selinux_label: Optional[str] = None
    uid: int
    signal_info: SignalInfo
    cause: Optional[str] = None
    causes: Optional[List[dict]] = None
    abort_message: Optional[str] = None
    extra: Optional[str] = None


class TombstoneCrashArtifact(AndroidArtifact):
    """
    Parser for Android tombstone crash files.

    This parser can parse both text and protobuf tombstone crash files.
    """

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["timestamp"],
            "module": self.__class__.__name__,
            "event": "Tombstone",
            "data": (
                f"Crash in '{record['process_name']}' process running as UID '{record['uid']}' in file '{record['file_name']}' "
                f"Crash type '{record['signal_info']['name']}' with code '{record['signal_info']['code_name']}'"
            ),
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_process(result["process_name"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )
                continue

            if result.get("command_line", []):
                command_name = result["command_line"][0].split("/")[-1]
                ioc_match = self.indicators.check_process(command_name)
                if ioc_match:
                    self.alertstore.critical(
                        ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                    )
                    continue

            SUSPICIOUS_UIDS = [
                0,  # root
                1000,  # system
                2000,  # shell
            ]
            if result["uid"] in SUSPICIOUS_UIDS:
                self.alertstore.medium(
                    (
                        f"Potentially suspicious crash in process '{result['process_name']}' "
                        f"running as UID '{result['uid']}' in tombstone '{result['file_name']}' at {result['timestamp']}"
                    ),
                    "",
                    result,
                )

    def parse_protobuf(
        self, file_name: str, file_timestamp: datetime.datetime, data: bytes
    ) -> None:
        """Parse Android tombstone crash files from a protobuf object."""
        self.results.append(self.parse_protobuf_record(file_name, file_timestamp, data))

    def parse_protobuf_record(
        self, file_name: str, file_timestamp: datetime.datetime, data: bytes
    ) -> dict:
        if not data:
            raise ValueError("empty protobuf tombstone")
        tombstone_pb = Tombstone().parse(data)
        tombstone_dict = tombstone_pb.to_dict(
            casing=betterproto2.Casing.SNAKE, include_default_values=True
        )

        # Add some extra metadata
        tombstone_dict["timestamp"] = self._parse_timestamp_string(
            tombstone_pb.timestamp
        )
        tombstone_dict["file_name"] = file_name
        tombstone_dict["file_timestamp"] = convert_datetime_to_iso(file_timestamp)
        tombstone_dict["process_name"] = self._proccess_name_from_thread(tombstone_dict)
        if isinstance(tombstone_dict.get("selinux_label"), str):
            tombstone_dict["selinux_label"] = tombstone_dict["selinux_label"].rstrip(
                "\x00"
            )

        # Confirm the tombstone is valid, and matches the output model
        tombstone = TombstoneCrashResult.model_validate(tombstone_dict)
        return tombstone.model_dump()

    def parse(
        self, file_name: str, file_timestamp: datetime.datetime, content: bytes
    ) -> None:
        """Parse text Android tombstone crash files."""
        self.results.append(self.parse_text_record(file_name, file_timestamp, content))

    def parse_text_record(
        self, file_name: str, file_timestamp: datetime.datetime, content: bytes
    ) -> dict:
        if not content:
            raise ValueError("empty plaintext tombstone")
        tombstone_dict = {
            "file_name": file_name,
            "file_timestamp": convert_datetime_to_iso(file_timestamp),
        }
        lines = content.decode("utf-8", errors="replace").splitlines()
        for line_num, line in enumerate(lines, 1):
            if not line.strip() or TOMBSTONE_DELIMITER in line:
                continue
            try:
                for key, destination_key in TOMBSTONE_TEXT_KEY_MAPPINGS.items():
                    if self._parse_tombstone_line(
                        line, key, destination_key, tombstone_dict
                    ):
                        break
            except Exception as e:
                raise ValueError(f"Error parsing line {line_num}: {str(e)}")

        # Validate the tombstone and add it to the results
        tombstone = TombstoneCrashResult.model_validate(tombstone_dict)
        return tombstone.model_dump()

    def _parse_tombstone_line(
        self, line: str, key: str, destination_key: str, tombstone: dict
    ) -> bool:
        if not line.startswith(f"{key}"):
            return False

        if key == "pid":
            return self._load_pid_line(line, tombstone)
        elif key == "signal":
            return self._load_signal_line(line, tombstone)
        elif key == "Timestamp":
            return self._load_timestamp_line(line, tombstone)
        else:
            return self._load_key_value_line(line, key, destination_key, tombstone)

    def _load_key_value_line(
        self, line: str, key: str, destination_key: str, tombstone: dict
    ) -> bool:
        # The caller matched the key as a bare prefix, so a longer word starting
        # with it arrives here: `Caused by: …` inside an abort message reaches
        # the `Cause` key. That is a different line, not a broken file — say so
        # by declining it, and let the remaining keys have their turn. Raising
        # here discarded the whole tombstone, crash and stack trace included.
        line_key, separator, value = line.partition(":")
        if not separator or line_key != key:
            return False

        value_clean = value.strip()
        if len(value_clean) >= 2 and value_clean[0] == value_clean[-1] == "'":
            value_clean = value_clean[1:-1]
        if destination_key == "uid":
            tombstone[destination_key] = int(value_clean)
        elif destination_key == "process_uptime":
            # eg. "Process uptime: 40s"
            tombstone[destination_key] = int(value_clean.rstrip("s"))
        elif destination_key == "command_line":
            # Wrap in list for consistency with protobuf format (repeated string).
            tombstone[destination_key] = [value_clean]
        else:
            tombstone[destination_key] = value_clean
        return True

    def _load_pid_line(self, line: str, tombstone: dict) -> bool:
        # The first pid line identifies the crashing thread. Full text tombstones
        # contain additional pid lines for the other threads in the process.
        if "pid" in tombstone:
            return True

        try:
            parts = line.split(" >>> ") if " >>> " in line else line.split(">>>")
            process_info = parts[0]

            # Parse pid, tid, name from process info
            info_parts = [p.strip() for p in process_info.split(",")]
            for info in info_parts:
                key, value = info.split(":", 1)
                key = key.strip()
                value = value.strip()

                if key == "pid":
                    tombstone["pid"] = int(value)
                elif key == "tid":
                    tombstone["tid"] = int(value)
                elif key == "name":
                    tombstone["process_name"] = value

            # Extract binary path if it exists
            if len(parts) > 1:
                tombstone["binary_path"] = parts[1].strip().rstrip(" <")

            return True

        except Exception as e:
            raise ValueError(f"Failed to parse PID line: {str(e)}")

    def _load_signal_line(self, line: str, tombstone: dict) -> bool:
        signal_part, code_part = map(str.strip, line.split(",")[:2])

        def parse_part(part: str, prefix: str) -> tuple[int, str]:
            match = part.split(prefix)[1]
            number = int(match.split()[0])
            name = match.split("(")[1].split(")")[0] if "(" in match else "UNKNOWN"
            return number, name

        signal_number, signal_name = parse_part(signal_part, "signal ")
        code_number, code_name = parse_part(code_part, "code ")

        tombstone["signal_info"] = {
            "code": code_number,
            "code_name": code_name,
            "name": signal_name,
            "number": signal_number,
        }
        return True

    def _load_timestamp_line(self, line: str, tombstone: dict) -> bool:
        timestamp = line.split(":", 1)[1].strip()
        tombstone["timestamp"] = self._parse_timestamp_string(timestamp)
        return True

    @staticmethod
    def _parse_timestamp_string(timestamp: str) -> str:
        timestamp_parsed = parser.parse(timestamp)
        return convert_datetime_to_iso(timestamp_parsed)

    @staticmethod
    def _proccess_name_from_thread(tombstone_dict: dict) -> str:
        if tombstone_dict.get("threads"):
            for thread in tombstone_dict["threads"].values():
                if thread.get("id") == tombstone_dict["tid"] and thread.get("name"):
                    return thread["name"]
        return "Unknown"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tombstone.proto
# plugin: python-betterproto2
from dataclasses import dataclass

import betterproto2


class Architecture(betterproto2.Enum):
    ARM32 = 0
    ARM64 = 1
    X86 = 2
    X86_64 = 3
    RISCV64 = 4
    NONE = 5


class MemoryErrorTool(betterproto2.Enum):
    GWP_ASAN = 0
    SCUDO = 1


class MemoryErrorType(betterproto2.Enum):
    UNKNOWN = 0
    USE_AFTER_FREE = 1
    DOUBLE_FREE = 2
    INVALID_FREE = 3
    BUFFER_OVERFLOW = 4
    BUFFER_UNDERFLOW = 5


@dataclass(eq=False, repr=False)
class CrashDetail(betterproto2.Message):
    """
    NOTE TO OEMS: If you add custom fields to this proto, do not use numbers in
    the reserved range.
    """

    name: "bytes" = betterproto2.field(1, betterproto2.TYPE_BYTES)
    data: "bytes" = betterproto2.field(2, betterproto2.TYPE_BYTES)


@dataclass(eq=False, repr=False)
class StackHistoryBufferEntry(betterproto2.Message):
    addr: "BacktraceFrame | None" = betterproto2.field(1, betterproto2.TYPE_MESSAGE, optional=True)
    fp: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)
    tag: "int" = betterproto2.field(3, betterproto2.TYPE_UINT64)


@dataclass(eq=False, repr=False)
class StackHistoryBuffer(betterproto2.Message):
    tid: "int" = betterproto2.field(1, betterproto2.TYPE_UINT64)
    entries: "list[StackHistoryBufferEntry]" = betterproto2.field(2, betterproto2.TYPE_MESSAGE, repeated=True)


@dataclass(eq=False, repr=False)
class Tombstone(betterproto2.Message):
    arch: "Architecture" = betterproto2.field(1, betterproto2.TYPE_ENUM, default_factory=lambda: Architecture(0))
    guest_arch: "Architecture" = betterproto2.field(24, betterproto2.TYPE_ENUM, default_factory=lambda: Architecture(0))
    build_fingerprint: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    revision: "str" = betterproto2.field(3, betterproto2.TYPE_STRING)
    timestamp: "str" = betterproto2.field(4, betterproto2.TYPE_STRING)
    pid: "int" = betterproto2.field(5, betterproto2.TYPE_UINT32)
    tid: "int" = betterproto2.field(6, betterproto2.TYPE_UINT32)
    uid: "int" = betterproto2.field(7, betterproto2.TYPE_UINT32)
    selinux_label: "str" = betterproto2.field(8, betterproto2.TYPE_STRING)
    command_line: "list[str]" = betterproto2.field(9, betterproto2.TYPE_STRING, repeated=True)
    # Process uptime in seconds.
    process_uptime: "int" = betterproto2.field(20, betterproto2.TYPE_UINT32)
    signal_info: "Signal | None" = betterproto2.field(10, betterproto2.TYPE_MESSAGE, optional=True)
    abort_message: "str" = betterproto2.field(14, betterproto2.TYPE_STRING)
    crash_details: "list[CrashDetail]" = betterproto2.field(21, betterproto2.TYPE_MESSAGE, repeated=True)
    causes: "list[Cause]" = betterproto2.field(15, betterproto2.TYPE_MESSAGE, repeated=True)
    threads: "dict[int, Thread]" = betterproto2.field(
        16, betterproto2.TYPE_MAP, map_meta=betterproto2.map_meta(betterproto2.TYPE_UINT32, betterproto2.TYPE_MESSAGE)
    )
    guest_threads: "dict[int, Thread]" = betterproto2.field(
        25, betterproto2.TYPE_MAP, map_meta=betterproto2.map_meta(betterproto2.TYPE_UINT32, betterproto2.TYPE_MESSAGE)
    )
    memory_mappings: "list[MemoryMapping]" = betterproto2.field(17, betterproto2.TYPE_MESSAGE, repeated=True)
    log_buffers: "list[LogBuffer]" = betterproto2.field(18, betterproto2.TYPE_MESSAGE, repeated=True)
    open_fds: "list[FD]" = betterproto2.field(19, betterproto2.TYPE_MESSAGE, repeated=True)
    page_size: "int" = betterproto2.field(22, betterproto2.TYPE_UINT32)
    has_been_16kb_mode: "bool" = betterproto2.field(23, betterproto2.TYPE_BOOL)
    stack_history_buffer: "StackHistoryBuffer | None" = betterproto2.field(26, betterproto2.TYPE_MESSAGE, optional=True)


@dataclass(eq=False, repr=False)
class Signal(betterproto2.Message):
    number: "int" = betterproto2.field(1, betterproto2.TYPE_INT32)
    name: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    code: "int" = betterproto2.field(3, betterproto2.TYPE_INT32)
    code_name: "str" = betterproto2.field(4, betterproto2.TYPE_STRING)
    has_sender: "bool" = betterproto2.field(5, betterproto2.TYPE_BOOL)
    sender_uid: "int" = betterproto2.field(6, betterproto2.TYPE_INT32)
    sender_pid: "int" = betterproto2.field(7, betterproto2.TYPE_INT32)
    has_fault_address: "bool" = betterproto2.field(8, betterproto2.TYPE_BOOL)
    fault_address: "int" = betterproto2.field(9, betterproto2.TYPE_UINT64)
    # Note, may or may not contain the dump of the actual memory contents.
    # Currently, on arm64, we only include metadata, and not the contents.
    fault_adjacent_metadata: "MemoryDump | None" = betterproto2.field(10, betterproto2.TYPE_MESSAGE, optional=True)


@dataclass(eq=False, repr=False)
class HeapObject(betterproto2.Message):
    address: "int" = betterproto2.field(1, betterproto2.TYPE_UINT64)
    size: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)
    allocation_tid: "int" = betterproto2.field(3, betterproto2.TYPE_UINT64)
    allocation_backtrace: "list[BacktraceFrame]" = betterproto2.field(4, betterproto2.TYPE_MESSAGE, repeated=True)
    deallocation_tid: "int" = betterproto2.field(5, betterproto2.TYPE_UINT64)
    deallocation_backtrace: "list[BacktraceFrame]" = betterproto2.field(6, betterproto2.TYPE_MESSAGE, repeated=True)


@dataclass(eq=False, repr=False)
class MemoryError(betterproto2.Message):
    tool: "MemoryErrorTool" = betterproto2.field(1, betterproto2.TYPE_ENUM, default_factory=lambda: MemoryErrorTool(0))
    type: "MemoryErrorType" = betterproto2.field(2, betterproto2.TYPE_ENUM, default_factory=lambda: MemoryErrorType(0))
    heap: "HeapObject | None" = betterproto2.field(3, betterproto2.TYPE_MESSAGE, optional=True, group="location")


@dataclass(eq=False, repr=False)
class Cause(betterproto2.Message):
    human_readable: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    memory_error: "MemoryError | None" = betterproto2.field(2, betterproto2.TYPE_MESSAGE, optional=True, group="details")


@dataclass(eq=False, repr=False)
class Register(betterproto2.Message):
    name: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    u64: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)


@dataclass(eq=False, repr=False)
class Thread(betterproto2.Message):
    id: "int" = betterproto2.field(1, betterproto2.TYPE_INT32)
    name: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    registers: "list[Register]" = betterproto2.field(3, betterproto2.TYPE_MESSAGE, repeated=True)
    backtrace_note: "list[str]" = betterproto2.field(7, betterproto2.TYPE_STRING, repeated=True)
    unreadable_elf_files: "list[str]" = betterproto2.field(9, betterproto2.TYPE_STRING, repeated=True)
    current_backtrace: "list[BacktraceFrame]" = betterproto2.field(4, betterproto2.TYPE_MESSAGE, repeated=True)
    memory_dump: "list[MemoryDump]" = betterproto2.field(5, betterproto2.TYPE_MESSAGE, repeated=True)
    tagged_addr_ctrl: "int" = betterproto2.field(6, betterproto2.TYPE_INT64)
    pac_enabled_keys: "int" = betterproto2.field(8, betterproto2.TYPE_INT64)


@dataclass(eq=False, repr=False)
class BacktraceFrame(betterproto2.Message):
    rel_pc: "int" = betterproto2.field(1, betterproto2.TYPE_UINT64)
    pc: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)
    sp: "int" = betterproto2.field(3, betterproto2.TYPE_UINT64)
    function_name: "str" = betterproto2.field(4, betterproto2.TYPE_STRING)
    function_offset: "int" = betterproto2.field(5, betterproto2.TYPE_UINT64)
    file_name: "str" = betterproto2.field(6, betterproto2.TYPE_STRING)
    file_map_offset: "int" = betterproto2.field(7, betterproto2.TYPE_UINT64)
    build_id: "str" = betterproto2.field(8, betterproto2.TYPE_STRING)


@dataclass(eq=False, repr=False)
class ArmMTEMetadata(betterproto2.Message):
    # One memory tag per granule (e.g. every 16 bytes) of regular memory.
    memory_tags: "bytes" = betterproto2.field(1, betterproto2.TYPE_BYTES)


@dataclass(eq=False, repr=False)
class MemoryDump(betterproto2.Message):
    register_name: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    mapping_name: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    begin_address: "int" = betterproto2.field(3, betterproto2.TYPE_UINT64)
    memory: "bytes" = betterproto2.field(4, betterproto2.TYPE_BYTES)
    arm_mte_metadata: "ArmMTEMetadata | None" = betterproto2.field(6, betterproto2.TYPE_MESSAGE, optional=True, group="metadata")


@dataclass(eq=False, repr=False)
class MemoryMapping(betterproto2.Message):
    begin_address: "int" = betterproto2.field(1, betterproto2.TYPE_UINT64)
    end_address: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)
    offset: "int" = betterproto2.field(3, betterproto2.TYPE_UINT64)
    read: "bool" = betterproto2.field(4, betterproto2.TYPE_BOOL)
    write: "bool" = betterproto2.field(5, betterproto2.TYPE_BOOL)
    execute: "bool" = betterproto2.field(6, betterproto2.TYPE_BOOL)
    mapping_name: "str" = betterproto2.field(7, betterproto2.TYPE_STRING)
    build_id: "str" = betterproto2.field(8, betterproto2.TYPE_STRING)
    load_bias: "int" = betterproto2.field(9, betterproto2.TYPE_UINT64)


@dataclass(eq=False, repr=False)
class FD(betterproto2.Message):
    fd: "int" = betterproto2.field(1, betterproto2.TYPE_INT32)
    path: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    owner: "str" = betterproto2.field(3, betterproto2.TYPE_STRING)
    tag: "int" = betterproto2.field(4, betterproto2.TYPE_UINT64)


@dataclass(eq=False, repr=False)
class LogBuffer(betterproto2.Message):
    name: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    logs: "list[LogMessage]" = betterproto2.field(2, betterproto2.TYPE_MESSAGE, repeated=True)


@dataclass(eq=False, repr=False)
class LogMessage(betterproto2.Message):
    timestamp: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    pid: "int" = betterproto2.field(2, betterproto2.TYPE_UINT32)
    tid: "int" = betterproto2.field(3, betterproto2.TYPE_UINT32)
    priority: "int" = betterproto2.field(4, betterproto2.TYPE_UINT32)
    tag: "str" = betterproto2.field(5, betterproto2.TYPE_STRING)
    message: "str" = betterproto2.field(6, betterproto2.TYPE_STRING)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import re
from typing import Dict, List

from mvt.android.utils import warn_android_patch_level

from .artifact import AndroidArtifact

INTERESTING_PROPERTIES = [
    "gsm.sim.operator.alpha",
    "gsm.sim.operator.iso-country",
    "persist.sys.timezone",
    "ro.boot.serialno",
    "ro.build.version.sdk",
    "ro.build.version.security_patch",
    "ro.product.cpu.abi",
    "ro.product.locale",
    "ro.product.vendor.manufacturer",
    "ro.product.vendor.model",
    "ro.product.vendor.name",
]


class GetProp(AndroidArtifact):
    def parse(self, entry: str) -> None:
        self.results: List[Dict[str, str]] = []
        # A property value may span several lines: persist.sys.boot.reason.history
        # prints one boot per line. Matching the whole section instead of line by
        # line lets a value run to the first closing bracket that ends a line.
        rxp = re.compile(
            r"^[ \t]*\[([^]]+)\]: \[(.*?)\][ \t\r]*$",
            re.MULTILINE | re.DOTALL,
        )

        for name, value in rxp.findall(entry):
            self.results.append({"name": name, "value": value})

    def get_device_timezone(self) -> str | None:
        """
        Get the device timezone from the getprop results

        Used in other moduels to calculate the timezone offset
        """
        for entry in self.results:
            if entry["name"] == "persist.sys.timezone":
                return entry["value"]
        return None

    def check_indicators(self) -> None:
        for entry in self.results:
            if entry["name"] in INTERESTING_PROPERTIES:
                self.log.info("%s: %s", entry["name"], entry["value"])

            if entry["name"] == "ro.build.version.security_patch":
                warning_message = warn_android_patch_level(entry["value"], self.log)
                if isinstance(warning_message, str):
                    self.alertstore.medium(warning_message, "", entry)

        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_android_property_name(
                result.get("name", "")
            )
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Optional

from mvt.common.module import DatabaseNotFoundError
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

SHUTDOWN_LOG_PATH = [
    "private/var/db/diagnostics/shutdown.log",
    "private/var/db/diagnostics/shutdown.*.log",
]


class ShutdownLog(IOSExtraction):
    """This module extracts processes information from the shutdown log file."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "shutdown",
            "data": f"Client {record['client']} with PID {record['pid']} "
            "was running when the device was shut down",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_file_path(result["client"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )
                continue

            for ioc in self.indicators.get_iocs("processes"):
                parts = result["client"].split("/")
                if ioc.value in parts:
                    self.alertstore.critical(
                        f'Found mention of a known malicious process "{ioc.value}" in shutdown.log',
                        "",
                        result,
                        matched_indicator=ioc,
                    )
                    continue

    def process_shutdownlog(self, content):
        current_processes = []
        recent_processes = []
        times_delayed = 0
        delay = 0.0
        for line in content.splitlines():
            line = line.strip()

            if line.startswith("remaining client pid:"):
                current_processes.append(
                    {
                        "pid": line[line.find("pid: ") + 5 : line.find(" (")],
                        "client": line[line.find("(") + 1 : line.find(")")],
                        "delay": delay,
                        "times_delayed": times_delayed,
                    }
                )
            elif line.startswith("After "):
                # Consider the previous processes
                # End of the current processes
                for p in current_processes:
                    recent_processes.append(p)
                delay += float(line.split(" ")[1][:-2])
                times_delayed += 1
                current_processes = []
            elif line.startswith("SIGTERM: "):
                for p in current_processes:
                    recent_processes.append(p)

                try:
                    mac_timestamp = int(line[line.find("[") + 1 : line.find("]")])
                except ValueError:
                    try:
                        start = line.find(" @") + 2
                        mac_timestamp = int(line[start : start + 10])
                    except Exception:
                        mac_timestamp = 0

                isodate = convert_mactime_to_iso(mac_timestamp, from_2001=False)

                for process in recent_processes:
                    self.results.append(
                        {
                            "isodate": isodate,
                            "pid": process["pid"],
                            "client": process["client"],
                            "delay": process["delay"],
                            "times_delayed": process["times_delayed"],
                        }
                    )

                current_processes = []
                recent_processes = []
                times_delayed = 0
                delay = 0.0

        self.results = sorted(self.results, key=lambda entry: entry["isodate"])

    def run(self) -> None:
        if self.file_path:
            shutdown_log_paths = [self.file_path]
        else:
            shutdown_log_paths = sorted(
                self._get_fs_files_from_patterns(SHUTDOWN_LOG_PATH)
            )

        if not shutdown_log_paths:
            raise DatabaseNotFoundError("unable to find any shutdown log files")

        for shutdown_log_path in shutdown_log_paths:
            self.log.info("Found shutdown log at path: %s", shutdown_log_path)
            with open(shutdown_log_path, "r", encoding="utf-8") as handle:
                self.process_shutdownlog(handle.read())



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

SAFARI_FAVICON_ROOT_PATHS = [
    "private/var/mobile/Library/Image Cache/Favicons/Favicons.db",
    "private/var/mobile/Containers/Data/Application/*/Library/Image Cache/Favicons/Favicons.db",
]


class SafariFavicon(IOSExtraction):
    """This module extracts all Safari favicon records."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.results: list = [] if results is None else results

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "safari_favicon",
            "data": f"Safari favicon from {record['url']} with icon URL "
            f"{record['icon_url']} ({record['type']})",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_url(result["url"])
            if not ioc_match:
                ioc_match = self.indicators.check_url(result["icon_url"])

            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def _process_favicon_db(self, file_path):
        conn = self._open_sqlite_db(file_path)

        # Fetch valid icon cache.
        cur = conn.cursor()
        cur.execute(
            """
            SELECT
                page_url.url,
                icon_info.url,
                icon_info.timestamp
            FROM page_url
            JOIN icon_info ON page_url.uuid = icon_info.uuid
            ORDER BY icon_info.timestamp;
        """
        )

        for row in cur:
            self.results.append(
                {
                    "url": row[0],
                    "icon_url": row[1],
                    "timestamp": row[2],
                    "isodate": convert_mactime_to_iso(row[2]),
                    "type": "valid",
                    "safari_favicon_db_path": file_path,
                }
            )

        # Fetch icons from the rejected icons table.
        cur.execute(
            """
            SELECT
                page_url,
                icon_url,
                timestamp
            FROM rejected_resources ORDER BY timestamp;
        """
        )

        for row in cur:
            self.results.append(
                {
                    "url": row[0],
                    "icon_url": row[1],
                    "timestamp": row[2],
                    "isodate": convert_mactime_to_iso(row[2]),
                    "type": "rejected",
                    "safari_favicon_db_path": file_path,
                }
            )

        cur.close()
        conn.close()

    def run(self) -> None:
        for file_path in self._get_fs_files_from_patterns(SAFARI_FAVICON_ROOT_PATHS):
            self.log.info("Found Safari favicon cache database at path: %s", file_path)
            self._process_favicon_db(file_path)

        self.log.info("Extracted a total of %d favicon records", len(self.results))

        self.results = sorted(self.results, key=lambda x: x["isodate"])



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Any, Optional

from mvt.android.modules.backup.base import BackupModule
from mvt.android.parsers.backup import parse_sms_file
from mvt.common.module_types import ModuleResults
from mvt.common.utils import check_for_links


class SMS(BackupModule):
    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.results: list[dict[str, Any]] = []

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        messages = []
        url_batches = []
        for message in self.results:
            if "body" not in message:
                continue

            message_links = message.get("links", [])
            if message_links == []:
                message_links = check_for_links(message.get("text", ""))

            messages.append(message)
            url_batches.append(message_links)

        for message, ioc_match in zip(
            messages, self.indicators.check_url_batches(url_batches)
        ):
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", message, matched_indicator=ioc_match.ioc
                )

    def collect_url_results(self) -> None:
        for message in self.results:
            for url in message.get("links", []):
                self.add_url_result(url, message.get("isodate"), "sms")

    def run(self) -> None:
        sms_path = "apps/com.android.providers.telephony/d_f/*_sms_backup"
        for file in self._get_files_by_pattern(sms_path):
            self.log.debug("Processing SMS backup file at %s", file)
            data = self._get_file_content(file)
            self.results.extend(parse_sms_file(data))

        mms_path = "apps/com.android.providers.telephony/d_f/*_mms_backup"
        for file in self._get_files_by_pattern(mms_path):
            self.log.debug("Processing MMS backup file at %s", file)
            data = self._get_file_content(file)
            self.results.extend(parse_sms_file(data))

        self.log.info("Extracted a total of %d SMS & MMS messages", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import collections
import logging
import plistlib
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

IDSTATUSCACHE_BACKUP_IDS = [
    "6b97989189901ceaa4e5be9b7f05fb584120e27b",
]
IDSTATUSCACHE_ROOT_PATHS = [
    "private/var/mobile/Library/Preferences/com.apple.identityservices.idstatuscache.plist",
    "private/var/mobile/Library/IdentityServices/idstatuscache.plist",
]


class IDStatusCache(IOSExtraction):
    """Extracts Apple Authentication information from idstatuscache.plist"""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "lookup",
            "data": f"Lookup of {record['user']} within {record['package']} "
            f"(Status {record['idstatus']})",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            if result.get("user", "").startswith("mailto:"):
                email = result["user"][7:].strip("'")
                ioc_match = self.indicators.check_email(email)
                if ioc_match:
                    self.alertstore.critical(
                        ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                    )
                    continue

            if "\\x00\\x00" in result.get("user", ""):
                self.alertstore.high(
                    f"Found an ID Status Cache entry with suspicious patterns: {result.get('user')}",
                    "",
                    result,
                )

    def _extract_idstatuscache_entries(self, file_path):
        with open(file_path, "rb") as handle:
            file_plist = plistlib.load(handle)

        id_status_cache_entries = []
        for app in file_plist:
            if not isinstance(file_plist[app], dict):
                continue

            for entry in file_plist[app]:
                try:
                    lookup_date = file_plist[app][entry]["LookupDate"]
                    id_status = file_plist[app][entry]["IDStatus"]
                except KeyError:
                    continue

                id_status_cache_entries.append(
                    {
                        "package": app,
                        "user": entry.replace("\x00", "\\x00"),
                        "isodate": convert_mactime_to_iso(lookup_date),
                        "idstatus": id_status,
                    }
                )

        entry_counter = collections.Counter(
            [entry["user"] for entry in id_status_cache_entries]
        )
        for entry in id_status_cache_entries:
            # Add total count of occurrences to the status cache entry.
            entry["occurrences"] = entry_counter[entry["user"]]
            self.results.append(entry)

    def run(self) -> None:
        if self.is_backup:
            self._find_ios_database(backup_ids=IDSTATUSCACHE_BACKUP_IDS)
            self.log.info("Found IDStatusCache plist at path: %s", self.file_path)
            self._extract_idstatuscache_entries(self.file_path)
        elif self.is_fs_dump:
            for idstatuscache_path in self._get_fs_files_from_patterns(
                IDSTATUSCACHE_ROOT_PATHS
            ):
                self.file_path = idstatuscache_path
                self.log.info("Found IDStatusCache plist at path: %s", self.file_path)
                self._extract_idstatuscache_entries(self.file_path)

        self.log.info(
            "Extracted a total of %d ID Status Cache entries", len(self.results)
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.url import URL
from mvt.common.utils import convert_mactime_to_datetime, convert_mactime_to_iso

from ..base import IOSExtraction

# Safari profiles (iOS 17 and later) each keep their own History.db under
# Library/Safari/Profiles/<UUID>/, separate from the default profile's database.
SAFARI_HISTORY_BACKUP_RELPATHS = [
    "Library/Safari/History.db",
    "Library/Safari/Profiles/*/History.db",
]
SAFARI_HISTORY_ROOT_PATHS = [
    "private/var/mobile/Library/Safari/History.db",
    "private/var/mobile/Library/Safari/Profiles/*/History.db",
    "private/var/mobile/Containers/Data/Application/*/Library/Safari/History.db",
    "private/var/mobile/Containers/Data/Application/*/Library/Safari/Profiles/*/History.db",
]


class SafariHistory(IOSExtraction):
    """This module extracts all Safari visits and tries to detect potential
    network injection attacks.


    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "safari_history",
            "data": f"Safari visit to {record['url']} (ID: {record['id']}, "
            f"Visit ID: {record['visit_id']})",
        }

    def _find_injections(self):
        for result in self.results:
            # We presume injections only happen on HTTP visits.
            if not result["url"].lower().startswith("http://"):
                continue

            # If there is no destination, no redirect happened.
            if not result["redirect_destination"]:
                continue

            try:
                origin_domain = URL(result["url"]).domain
            except Exception:
                origin_domain = ""

            # We loop again through visits in order to find redirect record.
            for redirect in self.results:
                if redirect["safari_history_db"] != result["safari_history_db"]:
                    continue

                if redirect["visit_id"] != result["redirect_destination"]:
                    continue

                try:
                    redirect_domain = URL(redirect["url"]).domain
                except Exception:
                    redirect_domain = ""

                # If the redirect destination is the same domain as the origin,
                # it's most likely an HTTPS upgrade.
                if origin_domain == redirect_domain:
                    continue

                self.log.info(
                    'Found HTTP redirect to different domain: "%s" -> "%s"',
                    origin_domain,
                    redirect_domain,
                )

                redirect_time = convert_mactime_to_datetime(redirect["timestamp"])
                origin_time = convert_mactime_to_datetime(result["timestamp"])
                elapsed_time = redirect_time - origin_time
                elapsed_ms = elapsed_time.microseconds / 1000

                if elapsed_time.seconds == 0:
                    self.alertstore.medium(
                        f"Redirect took less than a second! ({elapsed_ms} milliseconds)",
                        convert_mactime_to_iso(result["timestamp"]),
                        result,
                    )

    def check_indicators(self) -> None:
        self._find_injections()

        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_url(result["url"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def _process_history_db(self, history_path):
        self._recover_sqlite_db_if_needed(history_path)
        conn = self._open_sqlite_db(history_path)
        cur = conn.cursor()
        cur.execute(
            """
            SELECT
                history_items.id,
                history_items.url,
                history_visits.id,
                history_visits.visit_time,
                history_visits.redirect_source,
                history_visits.redirect_destination
            FROM history_items
            JOIN history_visits ON history_visits.history_item = history_items.id
            ORDER BY history_visits.visit_time;
        """
        )

        for row in cur:
            self.results.append(
                {
                    "id": row[0],
                    "url": row[1],
                    "visit_id": row[2],
                    "timestamp": row[3],
                    "isodate": convert_mactime_to_iso(row[3]),
                    "redirect_source": row[4],
                    "redirect_destination": row[5],
                    "safari_history_db": os.path.relpath(
                        history_path, self.target_path
                    ),
                }
            )

        cur.close()
        conn.close()

    def run(self) -> None:
        if self.is_backup:
            for relative_path in SAFARI_HISTORY_BACKUP_RELPATHS:
                for history_file in self._get_backup_files_from_manifest(
                    relative_path=relative_path
                ):
                    history_path = self._get_backup_file_from_id(
                        history_file["file_id"]
                    )

                    if not history_path:
                        continue

                    self.log.info(
                        "Found Safari history database at path: %s", history_path
                    )

                    self._process_history_db(history_path)
        elif self.is_fs_dump:
            for history_path in self._get_fs_files_from_patterns(
                SAFARI_HISTORY_ROOT_PATHS
            ):
                self.log.info("Found Safari history database at path: %s", history_path)
                self._process_history_db(history_path)

        self.log.info("Extracted a total of %d history records", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Any, Optional

import requests

from .config import settings

log = logging.getLogger(__name__)


class VTNoKey(Exception):
    pass


class VTQuotaExceeded(Exception):
    pass


def virustotal_lookup(file_hash: str) -> Optional[dict[str, Any]]:
    if not settings.VT_API_KEY:
        raise VTNoKey(
            "No VirusTotal API key provided: to use VirusTotal lookups please set "
            "MVT_VT_API_KEY or VT_API_KEY in the MVT configuration file"
        )

    headers = {
        "User-Agent": "VirusTotal",
        "Content-Type": "application/json",
        "x-apikey": settings.VT_API_KEY,
    }
    res = requests.get(
        f"https://www.virustotal.com/api/v3/files/{file_hash}",
        headers=headers,
        timeout=settings.NETWORK_TIMEOUT,
    )

    if res.status_code == 200:
        report = res.json()
        return report["data"]

    if res.status_code == 404:
        log.info("Could not find results for file with hash %s", file_hash)
    elif res.status_code == 429:
        raise VTQuotaExceeded("You have exceeded the quota for your VirusTotal API key")
    else:
        raise RuntimeError(f"Unexpected response from VirusTotal: {res.status_code}")

    return None



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import io
import json
import tarfile
import zlib

from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from mvt.common.utils import check_for_links, convert_unix_to_iso

PBKDF2_KEY_SIZE = 32


class AndroidBackupParsingError(Exception):
    """Exception raised file parsing an android backup file"""


class AndroidBackupNotImplemented(AndroidBackupParsingError):
    pass


class InvalidBackupPassword(AndroidBackupParsingError):
    pass


def to_utf8_bytes(input_bytes):
    output = []
    for byte in input_bytes:
        if byte < ord(b"\x80"):
            output.append(byte)
        else:
            output.append(ord("\xef") | (byte >> 12))
            output.append(ord("\xbc") | ((byte >> 6) & ord("\x3f")))
            output.append(ord("\x80") | (byte & ord("\x3f")))
    return bytes(output)


def parse_ab_header(data):
    """
    Parse the header of an Android Backup file
    Returns a dict {'backup': True, 'compression': False,
        'encryption': "none", 'version': 4}
    """
    if data.startswith(b"ANDROID BACKUP"):
        try:
            [_, version, is_compressed, encryption, _] = data.split(b"\n", 4)
            return {
                "backup": True,
                "compression": (is_compressed == b"1"),
                "version": int(version),
                "encryption": encryption.decode("utf-8"),
            }
        except (UnicodeDecodeError, ValueError):
            pass

    return {"backup": False, "compression": None, "version": None, "encryption": None}


def decrypt_master_key(
    password,
    user_salt,
    user_iv,
    pbkdf2_rounds,
    master_key_blob,
    format_version,
    checksum_salt,
):
    """Generate AES key from user password uisng PBKDF2

    The backup master key is extracted from the master key blog after decryption.
    """
    # Derive key from password using PBKDF2.
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA1(), length=32, salt=user_salt, iterations=pbkdf2_rounds
    )
    key = kdf.derive(password.encode("utf-8"))

    # Decrypt master key blob.
    cipher = Cipher(algorithms.AES(key), modes.CBC(user_iv))
    decryptor = cipher.decryptor()
    try:
        decryted_master_key_blob = (
            decryptor.update(master_key_blob) + decryptor.finalize()
        )

        # Extract key and IV from decrypted blob.
        key_blob = io.BytesIO(decryted_master_key_blob)
        master_iv_length = ord(key_blob.read(1))
        master_iv = key_blob.read(master_iv_length)

        master_key_length = ord(key_blob.read(1))
        master_key = key_blob.read(master_key_length)

        master_key_checksum_length = ord(key_blob.read(1))
        master_key_checksum = key_blob.read(master_key_checksum_length)
    except TypeError as exc:
        raise InvalidBackupPassword() from exc

    # Handle quirky encoding of master key bytes in Android original Java crypto code.
    if format_version > 1:
        hmac_mk = to_utf8_bytes(master_key)
    else:
        hmac_mk = master_key

    # Derive checksum to confirm successful backup decryption.
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA1(), length=32, salt=checksum_salt, iterations=pbkdf2_rounds
    )
    calculated_checksum = kdf.derive(hmac_mk)

    if master_key_checksum != calculated_checksum:
        raise InvalidBackupPassword()

    return master_key, master_iv


def decrypt_backup_data(encrypted_backup, password, encryption_algo, format_version):
    """
    Generate encryption keyffrom password and do decryption

    """
    if encryption_algo != b"AES-256":
        raise AndroidBackupNotImplemented("Encryption Algorithm not implemented")

    if password is None:
        raise InvalidBackupPassword()

    try:
        [
            user_salt,
            checksum_salt,
            pbkdf2_rounds,
            user_iv,
            master_key_blob,
            encrypted_data,
        ] = encrypted_backup.split(b"\n", 5)

        user_salt = bytes.fromhex(user_salt.decode("utf-8"))
        checksum_salt = bytes.fromhex(checksum_salt.decode("utf-8"))
        pbkdf2_rounds = int(pbkdf2_rounds)
        user_iv = bytes.fromhex(user_iv.decode("utf-8"))
        master_key_blob = bytes.fromhex(master_key_blob.decode("utf-8"))
    except (UnicodeDecodeError, ValueError) as exc:
        raise AndroidBackupParsingError("Invalid encrypted backup header") from exc

    # Derive decryption master key from password.
    master_key, master_iv = decrypt_master_key(
        password=password,
        user_salt=user_salt,
        user_iv=user_iv,
        pbkdf2_rounds=pbkdf2_rounds,
        master_key_blob=master_key_blob,
        format_version=format_version,
        checksum_salt=checksum_salt,
    )

    # Decrypt and unpad backup data using derived key.
    cipher = Cipher(algorithms.AES(master_key), modes.CBC(master_iv))
    decryptor = cipher.decryptor()
    decrypted_tar = decryptor.update(encrypted_data) + decryptor.finalize()

    unpadder = padding.PKCS7(128).unpadder()
    return unpadder.update(decrypted_tar) + unpadder.finalize()


def parse_backup_file(data, password=None):
    """
    Parse an ab file, returns a tar file

    """
    if not data.startswith(b"ANDROID BACKUP"):
        raise AndroidBackupParsingError("Invalid file header")

    try:
        [_, version, is_compressed, encryption_algo, tar_data] = data.split(b"\n", 4)
        version = int(version)
        is_compressed = int(is_compressed)
    except ValueError as exc:
        raise AndroidBackupParsingError("Invalid file header") from exc

    if encryption_algo != b"none":
        tar_data = decrypt_backup_data(
            tar_data, password, encryption_algo, format_version=version
        )

    if is_compressed:
        try:
            tar_data = zlib.decompress(tar_data)
        except zlib.error as exc:
            raise AndroidBackupParsingError(
                "Impossible to decompress the backup file"
            ) from exc

    return tar_data


def parse_tar_for_sms(data):
    """
    Extract SMS from a tar backup archive
    Returns an array of SMS
    """
    dbytes = io.BytesIO(data)

    res = []
    with tarfile.open(fileobj=dbytes) as tar:
        for member in tar.getmembers():
            if member.name.startswith("apps/com.android.providers.telephony/d_f/") and (
                member.name.endswith("_sms_backup")
                or member.name.endswith("_mms_backup")
            ):
                dhandler = tar.extractfile(member)
                if not dhandler:
                    continue
                res.extend(parse_sms_file(dhandler.read()))

    return res


def parse_sms_file(data):
    """
    Parse an SMS or MMS file extracted from a backup
    Returns a list of SMS or MMS entries
    """
    res = []
    data = zlib.decompress(data)
    json_data = json.loads(data)

    for entry in json_data:
        # Adapt MMS format to SMS format.
        if "mms_body" in entry:
            entry["body"] = entry["mms_body"]
            entry.pop("mms_body")

        body = entry.get("body", None)
        message_links = None
        if body:
            message_links = check_for_links(entry["body"])

        entry["isodate"] = convert_unix_to_iso(int(entry["date"]) / 1000)
        entry["direction"] = "sent" if int(entry["date_sent"]) else "received"

        # Extract links from the body
        if message_links or entry["body"].strip() == "":
            entry["links"] = message_links
        res.append(entry)

    return res



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import json
import logging
import os
import shutil
import tarfile
from pathlib import Path, PurePosixPath
from tempfile import TemporaryDirectory
from typing import Any, Optional

from mvt.common.command import Command
from mvt.common.indicators import Indicators
from mvt.common.module import MVTModule

log = logging.getLogger(__name__)


class CmdIOSCheckSysdiagnose(Command):
    def __init__(
        self,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        ioc_files: Optional[list] = None,
        iocs: Optional[Indicators] = None,
        module_name: Optional[str] = None,
        serial: Optional[str] = None,
        module_options: Optional[dict] = None,
        hashes: bool = False,
        sub_command: bool = False,
        disable_version_check: bool = False,
        disable_indicator_check: bool = False,
        custom_modules: Optional[list[type[MVTModule]]] = None,
    ) -> None:
        super().__init__(
            target_path=target_path,
            results_path=results_path,
            ioc_files=ioc_files,
            iocs=iocs,
            module_name=module_name,
            serial=serial,
            module_options=module_options,
            hashes=hashes,
            sub_command=sub_command,
            log=log,
            disable_version_check=disable_version_check,
            disable_indicator_check=disable_indicator_check,
            custom_modules=custom_modules,
        )
        self.platform = "ios"
        self.name = "check-sysdiagnose"
        self.sysdiagnose_format: Optional[str] = None
        self.sysdiagnose_archive: Optional[tarfile.TarFile] = None
        self.sysdiagnose_files: list[str] = []
        self.ips_files: list[dict[str, Any]] = []
        self.temp_sysdiagnose_dir: Optional[TemporaryDirectory[str]] = None
        self.extracted_sysdiagnose_path: Optional[str] = None

    @staticmethod
    def _parse_bugtype_header(data: bytes) -> Optional[int]:
        try:
            header = json.loads(data.split(b"\n", 1)[0].decode("utf-8"))
            return int(header["bug_type"])
        except (json.JSONDecodeError, KeyError, UnicodeDecodeError, ValueError):
            return None

    def _add_ips_file(self, file_path: str, data: bytes) -> None:
        bug_type = self._parse_bugtype_header(data)
        if bug_type is not None:
            self.ips_files.append({"file_path": file_path, "bug_type": bug_type})

    def init(self) -> None:
        if not self.target_path:
            raise ValueError("A sysdiagnose path is required")

        if os.path.isdir(self.target_path):
            self.sysdiagnose_format = "dir"
            parent_path = Path(self.target_path).absolute().parent
            for root, _, filenames in os.walk(self.target_path):
                for filename in filenames:
                    absolute_path = os.path.join(root, filename)
                    file_path = os.path.relpath(absolute_path, parent_path)
                    self.sysdiagnose_files.append(file_path)
                    if filename.endswith(".ips"):
                        with open(absolute_path, "rb") as handle:
                            self._add_ips_file(absolute_path, handle.read())
            return

        if not os.path.isfile(self.target_path):
            raise ValueError(f"Sysdiagnose path does not exist: {self.target_path}")

        self.log.info("Parsing sysdiagnose archive. This might take a while...")
        self.sysdiagnose_format = "tar"
        self.sysdiagnose_archive = tarfile.open(self.target_path, "r:gz")
        self._extract_sysdiagnose_archive()

    def _extract_sysdiagnose_archive(self) -> None:
        archive = self.sysdiagnose_archive
        if archive is None:
            raise RuntimeError("Sysdiagnose archive has not been initialized")

        self.temp_sysdiagnose_dir = TemporaryDirectory()
        extraction_root = Path(self.temp_sysdiagnose_dir.name).resolve()
        archive_roots = set()

        for member in archive:
            member_path = PurePosixPath(member.name.replace("\\", "/"))
            if member_path.is_absolute() or ".." in member_path.parts:
                self.log.warning("Skipping unsafe sysdiagnose path %r", member.name)
                continue

            destination = extraction_root.joinpath(*member_path.parts).resolve()
            if not destination.is_relative_to(extraction_root):
                self.log.warning("Skipping unsafe sysdiagnose path %r", member.name)
                continue

            if not member_path.parts:
                continue
            archive_roots.add(member_path.parts[0])

            if member.isdir():
                destination.mkdir(parents=True, exist_ok=True)
                continue

            # Modules only need directories and regular files. Do not materialize
            # links or device nodes from an untrusted sysdiagnose archive.
            if not member.isfile():
                self.log.warning("Skipping unsafe sysdiagnose member %r", member.name)
                continue

            normalized_name = member_path.as_posix()
            self.sysdiagnose_files.append(normalized_name)

            source = archive.extractfile(member)
            if source is None:
                continue

            destination.parent.mkdir(parents=True, exist_ok=True)
            with source, destination.open("wb") as output:
                shutil.copyfileobj(source, output)

            if normalized_name.endswith(".ips"):
                self._add_ips_file(str(destination), destination.read_bytes())

        if len(archive_roots) != 1:
            raise ValueError("Sysdiagnose archive must contain one top-level directory")

        self.extracted_sysdiagnose_path = str(extraction_root / archive_roots.pop())

    def module_init(self, module) -> None:
        module.ips_files = self.ips_files
        if self.sysdiagnose_format == "tar":
            if self.extracted_sysdiagnose_path is None:
                raise RuntimeError("Sysdiagnose archive has not been extracted")
            module.from_sysdiagnose_folder(
                self.extracted_sysdiagnose_path, self.sysdiagnose_files
            )
            return
        if self.sysdiagnose_format == "dir" and self.target_path:
            module.from_sysdiagnose_folder(self.target_path, self.sysdiagnose_files)
            return
        raise RuntimeError("Sysdiagnose input has not been initialized")

    def finish(self) -> None:
        if self.sysdiagnose_archive is not None:
            self.sysdiagnose_archive.close()
            self.sysdiagnose_archive = None
        if self.temp_sysdiagnose_dir is not None:
            self.temp_sysdiagnose_dir.cleanup()
            self.temp_sysdiagnose_dir = None



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_chrometime_to_datetime, convert_datetime_to_iso

from ..base import IOSExtraction

CHROME_HISTORY_BACKUP_IDS = [
    "faf971ce92c3ac508c018dce1bef2a8b8e9838f1",
]
CHROME_HISTORY_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/Library/Application Support/Google/Chrome/Default/History",  # pylint: disable=line-too-long
]


class ChromeHistory(IOSExtraction):
    """This module extracts all Chome visits."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "visit",
            "data": f"{record['id']} - {record['url']} "
            f"(visit ID: {record['visit_id']}, "
            f"redirect source: {record['redirect_source']})",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_url(result["url"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def run(self) -> None:
        self._find_ios_database(
            backup_ids=CHROME_HISTORY_BACKUP_IDS, root_paths=CHROME_HISTORY_ROOT_PATHS
        )
        self.log.info("Found Chrome history database at path: %s", self.file_path)

        if not self.file_path:
            return
        conn = self._open_sqlite_db(self.file_path)
        cur = conn.cursor()
        cur.execute(
            """
            SELECT
                urls.id,
                urls.url,
                visits.id,
                visits.visit_time,
                visits.from_visit
            FROM urls
            JOIN visits ON visits.url = urls.id
            ORDER BY visits.visit_time;
        """
        )

        for item in cur:
            self.results.append(
                {
                    "id": item[0],
                    "url": item[1],
                    "visit_id": item[2],
                    "timestamp": item[3],
                    "isodate": convert_datetime_to_iso(
                        convert_chrometime_to_datetime(item[3])
                    ),
                    "redirect_source": item[4],
                }
            )

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d history items", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
import plistlib
from base64 import b64encode
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_datetime_to_iso

from ..base import IOSExtraction

CONF_PROFILES_DOMAIN = (
    "SysSharedContainerDomain-systemgroup.com.apple.configurationprofiles"
)


class ConfigurationProfiles(IOSExtraction):
    """This module extracts the full plist data from configuration profiles."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        if not record["install_date"]:
            return {}

        payload_name = record["plist"].get("PayloadDisplayName")
        payload_description = record["plist"].get("PayloadDescription")
        return {
            "timestamp": record["install_date"],
            "module": self.__class__.__name__,
            "event": "configuration_profile_install",
            "data": f"{record['plist']['PayloadType']} installed: {record['plist']['PayloadUUID']} "
            f"- {payload_name}: {payload_description}",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            if result["plist"].get("PayloadUUID"):
                payload_content = result["plist"]["PayloadContent"][0]

                # Alert on any known malicious configuration profiles in the
                # indicator list.
                ioc_match = self.indicators.check_profile(
                    result["plist"]["PayloadUUID"]
                )
                if ioc_match:
                    warning_message = f'Found a known malicious configuration profile "{result["plist"]["PayloadDisplayName"]}" with UUID "{result["plist"]["PayloadUUID"]}"'
                    self.alertstore.critical(
                        warning_message, "", result, matched_indicator=ioc_match.ioc
                    )
                    continue

                # Highlight suspicious configuration profiles which may be used
                # to hide notifications.
                if payload_content["PayloadType"] in ["com.apple.notificationsettings"]:
                    warning_message = f'Found a potentially suspicious configuration profile "{result["plist"]["PayloadDisplayName"]}" with payload type {payload_content["PayloadType"]}'
                    self.alertstore.medium(warning_message, "", result)
                    continue

    @staticmethod
    def _b64encode_key(d: dict, key: str) -> None:
        if key in d:
            d[key] = b64encode(d[key])

    @staticmethod
    def _b64encode_keys(d: dict, keys: list) -> None:
        for key in keys:
            if key in d:
                d[key] = b64encode(d[key])

    def _b64encode_plist_bytes(self, plist: dict) -> None:
        """Encode binary plist values to base64 for JSON serialization."""
        if "SignerCerts" in plist:
            plist["SignerCerts"] = [b64encode(x) for x in plist["SignerCerts"]]

        self._b64encode_keys(plist, ["PushTokenDataSentToServerKey", "LastPushTokenHash"])

        if "OTAProfileStub" in plist:
            stub = plist["OTAProfileStub"]
            if "SignerCerts" in stub:
                stub["SignerCerts"] = [b64encode(x) for x in stub["SignerCerts"]]
            if "PayloadContent" in stub:
                self._b64encode_key(stub["PayloadContent"], "EnrollmentIdentityPersistentID")

        if "PayloadContent" in plist:
            for entry in plist["PayloadContent"]:
                self._b64encode_keys(entry, ["PERSISTENT_REF", "IdentityPersistentRef"])

    def run(self) -> None:
        for conf_file in self._get_backup_files_from_manifest(
            domain=CONF_PROFILES_DOMAIN
        ):
            conf_rel_path = conf_file["relative_path"]

            # Filter out all configuration files that are not configuration
            # profiles.
            if not conf_rel_path or not os.path.basename(conf_rel_path).startswith(
                "profile-"
            ):
                continue

            conf_file_path = self._get_backup_file_from_id(conf_file["file_id"])
            if not conf_file_path:
                self.log.debug(
                    "Missing file %s in backup (%s)",
                    conf_file["file_id"],
                    conf_file["relative_path"],
                )
                continue

            with open(conf_file_path, "rb") as handle:
                try:
                    conf_plist = plistlib.load(handle)
                except Exception:
                    conf_plist = {}

            self._b64encode_plist_bytes(conf_plist)

            self.results.append(
                {
                    "file_id": conf_file["file_id"],
                    "relative_path": conf_file["relative_path"],
                    "domain": conf_file["domain"],
                    "plist": conf_plist,
                    "install_date": convert_datetime_to_iso(
                        conf_plist.get("InstallDate")
                    ),
                }
            )

        self.log.info(
            "Extracted details about %d configuration profiles", len(self.results)
        )



import json
import os
from typing import Optional, Tuple, Type

import yaml
from appdirs import user_config_dir
from pydantic import Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)

MVT_CONFIG_FOLDER = user_config_dir("mvt")
MVT_CONFIG_PATH = os.path.join(MVT_CONFIG_FOLDER, "config.yaml")


class MVTSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="MVT_",
        env_nested_delimiter="_",
        extra="ignore",
    )
    # Allow to decided if want to load environment variables
    load_env: bool = Field(True, exclude=True)

    # General settings
    PYPI_UPDATE_URL: str = Field(
        default="https://pypi.org/pypi/mvt/json",
    )
    NETWORK_ACCESS_ALLOWED: bool = True
    NETWORK_TIMEOUT: int = 15

    # Command default settings, all can be specified by MVT_ prefixed environment variables too.
    IOS_BACKUP_PASSWORD: Optional[str] = Field(
        default=None, description="Default password to use to decrypt iOS backups"
    )
    ANDROID_BACKUP_PASSWORD: Optional[str] = Field(
        default=None, description="Default password to use to decrypt Android backups"
    )
    STIX2: Optional[str] = Field(
        default=None, description="List of directories where STIX2 files are stored"
    )
    VT_API_KEY: Optional[str] = Field(
        default=None, description="API key to use for VirusTotal lookups"
    )
    PROFILE: bool = Field(
        default=False, description="Profile the execution of MVT modules"
    )
    HASH_FILES: bool = Field(default=False, description="Should MVT hash output files")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        yaml_source = YamlConfigSettingsSource(settings_cls, MVT_CONFIG_PATH)
        sources: Tuple[PydanticBaseSettingsSource, ...] = (
            yaml_source,
            init_settings,
        )
        # Always load env variables by default
        sources = (env_settings,) + sources
        return sources

    def save_settings(
        self,
    ) -> None:
        """
        Save the current settings to a file.
        """
        if not os.path.isdir(MVT_CONFIG_FOLDER):
            os.makedirs(MVT_CONFIG_FOLDER)

        # Dump the settings to the YAML file
        model_serializable = json.loads(self.model_dump_json(exclude_defaults=True))
        with open(MVT_CONFIG_PATH, "w") as config_file:
            config_file.write(yaml.dump(model_serializable, default_flow_style=False))

    @classmethod
    def initialise(cls) -> "MVTSettings":
        """
        Initialise the settings file.

        We first initialise the settings (without env variable) and then persist
        them to file. This way we can update the config file with the default values.

        Afterwards we load the settings again, this time including the env variables.
        """
        # Set invalid env prefix to avoid loading env variables.
        settings = cls(load_env=False)
        settings.save_settings()

        # Load the settings again with any ENV variables.
        settings = cls(load_env=True)
        return settings


settings = MVTSettings.initialise()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import fnmatch
import logging
import os
from tarfile import TarFile
from typing import List, Optional

from mvt.common.module import ModuleResults, MVTModule


class BackupModule(MVTModule):
    """This class provides a base for all backup extractios modules"""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.ab: Optional[str] = None
        self.backup_path: Optional[str] = None
        self.tar: Optional[TarFile] = None
        self.files: list = []

    def from_dir(self, backup_path: Optional[str], files: List[str]) -> None:
        self.backup_path = backup_path
        self.files = files

    def from_ab(
        self, file_path: Optional[str], tar: Optional[TarFile], files: List[str]
    ) -> None:
        """
        Extract the files
        """
        self.ab = file_path
        self.tar = tar
        self.files = files

    def _get_files_by_pattern(self, pattern: str) -> list:
        return fnmatch.filter(self.files, pattern)

    def _get_file_content(self, file_path: str) -> bytes:
        handle = None
        if self.tar:
            try:
                member = self.tar.getmember(file_path)
                handle = self.tar.extractfile(member)
                if not handle:
                    raise ValueError(f"Could not extract file: {file_path}")
            except KeyError:
                raise FileNotFoundError(f"File not found in tar: {file_path}")
        elif self.backup_path:
            handle = open(os.path.join(self.backup_path, file_path), "rb")
        else:
            raise ValueError("No backup path or tar file provided")

        data = handle.read()
        handle.close()

        return data



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from mvt.common.module import MVTModule

from .aqf_files import AQFFiles
from .aqf_getprop import AQFGetProp
from .aqf_log_timestamps import AQFLogTimestamps
from .aqf_packages import AQFPackages
from .aqf_processes import AQFProcesses
from .aqf_settings import AQFSettings
from .mounts import Mounts
from .root_binaries import RootBinaries

ANDROIDQF_MODULES: list[type[MVTModule]] = [
    AQFPackages,
    AQFProcesses,
    AQFGetProp,
    AQFSettings,
    AQFFiles,
    AQFLogTimestamps,
    RootBinaries,
    Mounts,
]



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import fnmatch
import logging
import os
import zipfile
from typing import List, Optional

from mvt.common.module import MVTModule
from mvt.common.module_types import ModuleResults


class AndroidQFModule(MVTModule):
    """This class provides a base for all Android Data analysis modules."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.parent_path: Optional[str] = None
        self._path: Optional[str] = target_path
        self.files: List[str] = []
        self.archive: Optional[zipfile.ZipFile] = None

    def from_dir(self, parent_path: str, files: List[str]) -> None:
        self.parent_path = parent_path
        self.files = files

    def from_zip(self, archive: zipfile.ZipFile, files: List[str]) -> None:
        self.archive = archive
        self.files = files

    def _get_files_by_pattern(self, pattern: str):
        return fnmatch.filter(self.files, pattern)

    def _get_device_timezone(self):
        """
        Get the device timezone from the getprop.txt file.

        This is needed to map local timestamps stored in some
        Android log files to UTC/timezone-aware timestamps.
        """
        get_prop_files = self._get_files_by_pattern("*/getprop.txt")
        if not get_prop_files:
            self.log.warning(
                "Could not find getprop.txt file. "
                "Some timestamps and timeline data may be incorrect."
            )
            return None

        from mvt.android.artifacts.getprop import GetProp

        properties_artifact = GetProp()
        prop_data = self._get_file_content(get_prop_files[0]).decode("utf-8")
        properties_artifact.parse(prop_data)
        timezone = properties_artifact.get_device_timezone()
        if timezone:
            self.log.debug("Identified local phone timezone: %s", timezone)
            return timezone

        self.log.warning(
            "Could not find or determine local device timezone. "
            "Some timestamps and timeline data may be incorrect."
        )
        return None

    def _get_file_content(self, file_path):
        if self.archive:
            handle = self.archive.open(file_path)
        else:
            handle = open(os.path.join(self.parent_path, file_path), "rb")

        data = handle.read()
        handle.close()

        return data



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from mvt.common.module import MVTModule

from .backup_info import BackupInfo
from .configuration_profiles import ConfigurationProfiles
from .manifest import Manifest
from .profile_events import ProfileEvents

BACKUP_MODULES: list[type[MVTModule]] = [
    BackupInfo,
    ConfigurationProfiles,
    Manifest,
    ProfileEvents,
]
